# !/usr/bin/python
# -*- coding: latin-1 -*-
"""
A module for utility funtions
"""

import glob
import itertools
import os
import re
import requests
import shutil
import urllib

from astropy.io import fits
import bokeh.palettes as bpal
from scipy.interpolate import RegularGridInterpolator
import matplotlib.pyplot as plt
import numpy as np
from svo_filters import svo

# Supported profiles
PROFILES = ['uniform', 'linear', 'quadratic',
            'square-root', 'logarithmic', 'exponential',
            '3-parameter', '4-parameter']

# Supported filters
FILTERS = svo.filters()

# Get the location of EXOCTK_DATA environvment variable and check that it is valid
EXOCTK_DATA = os.environ.get('EXOCTK_DATA')

# If the variable is blank or doesn't exist
HOME_DIR = os.path.expanduser('~')
ON_TRAVIS_OR_RTD = HOME_DIR == '/home/travis' or HOME_DIR == '/Users/travis' or HOME_DIR == '/home/docs'
if not ON_TRAVIS_OR_RTD:
    if not EXOCTK_DATA:
        print(
            'WARNING: The $EXOCTK_DATA environment variable is not set.  Please set the '
            'value of this variable to point to the location of the exoctk_data '
            'download folder.  Users may retreive this folder by clicking the '
            '"ExoCTK Data Download" button on the ExoCTK website, or by using '
            'the exoctk.utils.download_exoctk_data() function.')
    else:
        # If the variable exists but doesn't point to a real location
        if not os.path.exists(EXOCTK_DATA):
            print(
                'WARNING: The $EXOCTK_DATA environment variable is set to a location that '
                'cannot be accessed.')

        # If the variable exists, points to a real location, but is missing contents
        for item in ['exoctk_contam', 'exoctk_log', 'fortney', 'generic', 'groups_integrations', 'modelgrid']:
            if item not in [os.path.basename(item) for item in glob.glob(os.path.join(EXOCTK_DATA, '*'))]:
                print(
                    'WARNING: Missing {}/ directory from {}. Please ensure that the ExoCTK data package has been '
                    'downloaded. Users may retrieve this package by clicking the "ExoCTK Data Download" '
                    'button on the ExoCTK website, or by using the exoctk.utils.download_exoctk_data() '
                    'function'.format(item, EXOCTK_DATA))

        EXOCTK_CONTAM_DIR = os.path.join(EXOCTK_DATA, 'exoctk_contam/')
        EXOCTKLOG_DIR = os.path.join(EXOCTK_DATA, 'exoctk_log/')
        FORTGRID_DIR = os.path.join(EXOCTK_DATA, 'fortney/')
        GENERICGRID_DIR = os.path.join(EXOCTK_DATA, 'generic/')
        GROUPS_INTEGRATIONS_DIR = os.path.join(EXOCTK_DATA, 'groups_integrations/')
        MODELGRID_DIR = os.path.join(EXOCTK_DATA, 'modelgrid/')


def download_exoctk_data(download_location=os.path.expanduser('~')):
    """Retrieves the ``exoctk_data`` materials from Box, downloads them
    to the user's local machine, uncompresses the files, and arranges
    them into an ``exoctk_data`` directory.

    Parameters
    ----------
    download_location : string
        The path to where the ExoCTK data package will be downloaded.
        The default setting is the user's $HOME directory.
    """

    print('\nDownloading ExoCTK data package.  This may take a few minutes.')
    print('Materials will be downloaded to {}/exoctk_data/\n'.format(download_location))

    # Ensure the exoctk_data/ directory exists in user's home directory
    exoctk_data_dir = os.path.join(download_location, 'exoctk_data')
    try:
        if not os.path.exists(exoctk_data_dir):
            os.makedirs(exoctk_data_dir)
    except PermissionError:
        print('Data download failed.  Unable to create {}.  Please check permissions.')

    # URLs to download contents
    urls = ['https://data.science.stsci.edu/redirect/JWST/ExoCTK/compressed/exoctk_contam.tar.gz',
            'https://data.science.stsci.edu/redirect/JWST/ExoCTK/compressed/exoctk_log.tar.gz',
            'https://data.science.stsci.edu/redirect/JWST/ExoCTK/compressed/groups_integrations.tar.gz',
            'https://data.science.stsci.edu/redirect/JWST/ExoCTK/compressed/fortney.tar.gz',
            'https://data.science.stsci.edu/redirect/JWST/ExoCTK/compressed/generic.tar.gz',
            'https://data.science.stsci.edu/redirect/JWST/ExoCTK/compressed/modelgrid_ATLAS9.tar.gz',
            'https://data.science.stsci.edu/redirect/JWST/ExoCTK/compressed/modelgrid_ACES_1.tar.gz',
            'https://data.science.stsci.edu/redirect/JWST/ExoCTK/compressed/modelgrid_ACES_2.tar.gz']

    # Build landing paths for downloads
    download_paths = [os.path.join(exoctk_data_dir, os.path.basename(url)) for url in urls]

    # Perform the downloads
    for i, url in enumerate(urls):
        landing_path = os.path.join(exoctk_data_dir, os.path.basename(url))
        print('({}/{}) Downloading data to {} from {}'.format(i+1, len(urls), landing_path, url))
        with requests.get(url, stream=True) as response:
            with open(landing_path, 'wb') as f:
                for chunk in response.iter_content(chunk_size=2048):
                    if chunk:
                        f.write(chunk)
    print('\nDownload complete\n')

    # Uncompress data
    print('Uncompressing data:\n')
    for path in download_paths:
        # Uncompress data
        print('\t{}'.format(path))
        shutil.unpack_archive(path, exoctk_data_dir)
        # Remove original .tar.gz files
        os.remove(path)

    # Combine modelgrid directories
    print('\nOrganizing files into exoctk_data/ directory')
    try:
        os.makedirs(os.path.join(exoctk_data_dir, 'modelgrid', 'ATLAS9'))
        os.makedirs(os.path.join(exoctk_data_dir, 'modelgrid', 'ACES'))
    except FileExistsError:
        pass
    modelgrid_files = glob.glob(os.path.join(exoctk_data_dir, 'modelgrid.*', '*'))
    for src in modelgrid_files:
        if 'ATLAS9' in src:
            dst = os.path.join(exoctk_data_dir, 'modelgrid', 'ATLAS9')
        elif 'ACES_' in src:
            dst = os.path.join(exoctk_data_dir, 'modelgrid', 'ACES')
        try:
            shutil.move(src, dst)
        except shutil.Error:
            print('Unable to organize modelgrid/ directory')
    shutil.rmtree(os.path.join(exoctk_data_dir, 'modelgrid.ATLAS9'))
    shutil.rmtree(os.path.join(exoctk_data_dir, 'modelgrid.ACES_1'))
    shutil.rmtree(os.path.join(exoctk_data_dir, 'modelgrid.ACES_2'))

    print('Completed!')


def color_gen(colormap='viridis', key=None, n=10):
    """Color generator for Bokeh plots

    Parameters
    ----------
    colormap: str, sequence
        The name of the color map

    Returns
    -------
    generator
        A generator for the color palette
    """
    if colormap in dir(bpal):
        palette = getattr(bpal, colormap)

        if isinstance(palette, dict):
            if key is None:
                key = list(palette.keys())[0]
            palette = palette[key]

        elif callable(palette):
            palette = palette(n)

        else:
            raise TypeError("pallette must be a bokeh palette name or a sequence of color hex values.")

    elif isinstance(colormap, (list, tuple)):
        palette = colormap

    else:
        raise TypeError("pallette must be a bokeh palette name or a sequence of color hex values.")

    yield from itertools.cycle(palette)


COLORS = color_gen('Category10')


def interp_flux(mu, flux, params, values):
    """
    Interpolate a cube of synthetic spectra for a
    given index of mu

    Parameters
    ----------
    mu: int
        The index of the (Teff, logg, FeH, *mu*, wavelength)
        data cube to interpolate
    flux: np.ndarray
        The 5D data array
    params: list
        A list of each free parameter range
    values: list
        A list of each free parameter values

    Returns
    -------
    tu
        The array of new flux values
    """
    # Iterate over each wavelength (-1 index of flux array)
    shp = flux.shape[-1]
    flx = np.zeros(shp)
    generators = []
    for lam in range(shp):
        interp_f = RegularGridInterpolator(params, flux[:, :, :, mu, lam])
        f, = interp_f(values)

        flx[lam] = f
        generators.append(interp_f)

    return flx, generators


def calc_zoom(R_f, arr):
    """
    Calculate the zoom factor required to make the given
    array into the given resolution

    Parameters
    ----------
    R_f: int
        The desired final resolution of the wavelength array
    arr: array-like
        The array to zoom
    """
    # Get initial resolution
    lam = arr[-1]-arr[0]
    d_lam_i = np.nanmean(np.diff(arr))
    # R_i = lam/d_lam_i

    # Calculate zoom
    d_lam_f = lam/R_f
    z = d_lam_i/d_lam_f

    return z


def rebin_spec(spec, wavnew, oversamp=100, plot=False):
    """
    Rebin a spectrum to a new wavelength array while preserving
    the total flux

    Parameters
    ----------
    spec: array-like
        The wavelength and flux to be binned
    wavenew: array-like
        The new wavelength array

    Returns
    -------
    np.ndarray
        The rebinned flux

    """
    wave, flux = spec
    nlam = len(wave)
    x0 = np.arange(nlam, dtype=float)
    x0int = np.arange((nlam-1.) * oversamp + 1., dtype=float)/oversamp
    w0int = np.interp(x0int, x0, wave)
    spec0int = np.interp(w0int, wave, flux)/oversamp

    # Set up the bin edges for down-binning
    maxdiffw1 = np.diff(wavnew).max()
    w1bins = np.concatenate(([wavnew[0]-maxdiffw1],
                             .5*(wavnew[1::]+wavnew[0: -1]),
                             [wavnew[-1]+maxdiffw1]))

    # Bin down the interpolated spectrum:
    w1bins = np.sort(w1bins)
    nbins = len(w1bins)-1
    specnew = np.zeros(nbins)
    inds2 = [[w0int.searchsorted(w1bins[ii], side='left'),
              w0int.searchsorted(w1bins[ii+1], side='left')]
             for ii in range(nbins)]

    for ii in range(nbins):
        specnew[ii] = np.sum(spec0int[inds2[ii][0]: inds2[ii][1]])

    return specnew


def writeFITS(filename, extensions, headers=()):
    """
    Write some data to a new FITS file

    Parameters
    ----------
    filename: str
        The filename of the output FITS file
    extensions: dict
        The extension name and associated data to include
        in the file
    headers: array-like
        The (keyword, value, comment) groups for the PRIMARY
        header extension

    """
    # Write the arrays to a FITS file
    prihdu = fits.PrimaryHDU()
    prihdu.name = 'PRIMARY'
    hdulist = fits.HDUList([prihdu])

    # Write the header to the PRIMARY HDU
    hdulist['PRIMARY'].header.extend(headers, end=True)

    # Write the data to the HDU
    for k, v in extensions.items():
        hdulist.append(fits.ImageHDU(data=v, name=k))

    # Write the file
    hdulist.writeto(filename, clobber=True)
    hdulist.close()

    # Insert END card to prevent header error
    # hdulist[0].header.tofile(filename, endcard=True, clobber=True)


def smooth(x, window_len=10, window='hanning'):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    Parameters
    ----------
    x: sequence
        The input signal
    window_len: int
        The dimension of the smoothing window
    window: str
        The type of window from 'flat', 'hanning', 'hamming', 'bartlett',
        'blackman'. 'flat' window will produce a moving average smoothing.

    Retruns
    -------
    np.ndarray
        The smoothed signal

    Example
    -------
    t = linspace(-2, 2, 0.1)
    x = sin(t)+randn(len(t))*0.1
    y = smooth(x)
    """
    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if window not in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is one of 'flat', 'hanning', 'hamming',\
                          'bartlett', 'blackman'")

    s = np.r_[2*np.median(x[0: window_len/5])-x[window_len: 1: -1], x,
              2*np.median(x[-window_len/5:])-x[-1: -window_len: -1]]

    if window == 'flat':
        w = np.ones(window_len, 'd')
    else:
        w = eval('np.'+window+'(window_len)')

    y = np.convolve(w/w.sum(), s, mode='same')

    return y[window_len-1: -window_len+1]


def medfilt(x, window_len):
    """
    Apply a length-k median filter to a 1D array x.
    Boundaries are extended by repeating endpoints.

    Parameters
    ----------
    x: np.array
        The 1D array to smooth
    window_len: int
        The size of the smoothing window

    Returns
    -------
    np.ndarray
        The smoothed 1D array
    """
    # assert x.ndim == 1, "Input must be one-dimensional."
    if window_len % 2 == 0:
        s1 = "Median filter length ("
        s2 = ") must be odd. Adding 1."
        print(s1+str(window_len)+s2)
        window_len += 1
    window_len = int(window_len)
    k2 = int((window_len - 1)//2)
    s = np.r_[2*np.median(x[0: int(window_len/5)])-x[window_len: 1: -1],
              x, 2*np.median(x[int(-window_len/5):])-x[-1: -window_len: -1]]
    y = np.zeros((len(s), window_len), dtype=s.dtype)

    y[:, k2] = s
    for i in range(k2):
        j = k2 - i
        y[j:, i] = s[:-j]
        y[: j, i] = s[0]
        y[: -j, -(i+1)] = s[j:]
        y[-j:, -(i+1)] = s[-1]
    return np.median(y[window_len-1: -window_len+1], axis=1)


def filter_table(table, **kwargs):
    """Retrieve the filtered rows

    Parameters
    ----------
    table: astropy.table.Table, pandas.DataFrame
        The table to filter
    param: str
        The parameter to filter by, e.g. 'Teff'
    value: str, float, int, sequence
        The criteria to filter by,
        which can be single valued like 1400
        or a range with operators [<,<=,>,>=],
        e.g. ('>1200','<=1400')

    Returns
    -------
    astropy.table.Table, pandas.DataFrame
        The filtered table
    """
    for param, value in kwargs.items():

        # Check it is a valid column
        if param not in table.colnames:
            raise KeyError("No column named {}".format(param))

        # Wildcard case
        if isinstance(value, (str, bytes)) and '*' in value:

            # Get column data
            data = list(map(str, table[param]))

            if not value.startswith('*'):
                value = '^'+value
            if not value.endswith('*'):
                value = value+'$'

            # Strip double quotes and decod
            value = value.replace("'", '')\
                         .replace('"', '')\
                         .replace('*', '(.*)')

            # Regex
            reg = re.compile(value, re.IGNORECASE)
            keep = list(filter(reg.findall, data))

            # Get indexes
            idx = np.where([i in keep for i in data])

            # Filter table
            table = table[idx]

        else:

            # Make single value string into conditions
            if isinstance(value, str):

                # Check for operator
                if any([value.startswith(o) for o in ['<', '>', '=']]):
                    value = [value]

                # Assume eqality if no operator
                else:
                    value = ['=='+value]

            # Turn numbers into strings
            if isinstance(value, (int, float, np.float16)):
                value = ["=={}".format(value)]

            # Iterate through multiple conditions
            for cond in value:

                # Equality
                if cond.startswith('='):
                    v = cond.replace('=', '')
                    if v.replace('.','',1).isdigit():
                        table = table[table[param] == eval(v)]
                    else:
                        table = table[table[param] == v]

                # Less than or equal
                elif cond.startswith('<='):
                    v = cond.replace('<=', '')
                    table = table[table[param] <= eval(v)]

                # Less than
                elif cond.startswith('<'):
                    v = cond.replace('<', '')
                    table = table[table[param] < eval(v)]

                # Greater than or equal
                elif cond.startswith('>='):
                    v = cond.replace('>=', '')
                    table = table[table[param] >= eval(v)]

                # Greater than
                elif cond.startswith('>'):
                    v = cond.replace('>', '')
                    table = table[table[param] > eval(v)]

                else:
                    raise ValueError("'{}' operator not valid.".format(cond))

    return table


def find_closest(axes, points, n=1, values=False):
    """
    Find the n-neighboring elements of a given value in an array

    Parameters
    ----------
    axes: list, np.array
        The array(s) to search
    points: array-like, float
        The point(s) to search for
    n: int
        The number of values to the left and right of the points
    Returns
    -------
    np.ndarray
        The n-values to the left and right of 'points' in 'axes'
    """
    results = []
    if not isinstance(axes, list):
        axes = [axes]
        points = [points]

    for i, (axis, point) in enumerate(zip(axes, points)):
        if point >= min(axis) and point <= max(axis):
            axis = np.asarray(axis)
            idx = np.clip(axis.searchsorted(point), 1, len(axis)-1)
            slc = slice(max(0, idx-n), min(idx+n, len(axis)))

            if values:
                result = axis[slc]
            else:

                result = np.arange(0, len(axis))[slc].astype(int)

            results.append(result)
        else:
            print('Point {} outside grid.'.format(point))
            return

    return results

def build_target_url(target_name):
    '''Build restful api url based on target name.

    Parameters
        ----------
        target_name : string
            The name of the target transit.

        Returns
        -------
        target_url : string
    '''
    # Encode the target name string.
    encode_target_name = urllib.parse.quote(target_name, encoding='utf-8')
    target_url = "https://exo.mast.stsci.edu/api/v0.1/exoplanets/{}/properties/".format(encode_target_name)

    return target_url

def get_canonical_name(target_name):
    '''Get ExoMAST prefered name for exoplanet.

        Parameters
        ----------
        target_name : string
            The name of the target transit.

        Returns
        -------
        canonical_name : string
    '''

    target_url = "https://exo.mast.stsci.edu/api/v0.1/exoplanets/identifiers/"

    # Create params dict for url parsing. Easier than trying to format yourself.
    params = {"name":target_name}

    r = requests.get(target_url, params=params)
    planetnames = r.json()
    canonical_name = planetnames['canonicalName']

    return canonical_name

def get_env_variables():
    """Returns a dictionary containing various environment variable
    information.

    Returns
    -------
    env_variables : dict
        A dictionary containing various environment variable data
    """

    env_variables = {}

    # Get the location of EXOCTK_DATA environvment variable and check that it is valid
    env_variables['exoctk_data'] = os.environ.get('EXOCTK_DATA')

    # If the variable is blank or doesn't exist
    ON_TRAVIS = os.path.expanduser('~') == '/home/travis' or os.path.expanduser('~') == '/Users/travis'
    if not ON_TRAVIS:
        if not env_variables['exoctk_data']:
            raise ValueError(
                'The $EXOCTK_DATA environment variable is not set.  Please set the '
                'value of this variable to point to the location of the ExoCTK data '
                'download folder.  Users may retreive this folder by clicking the '
                '"ExoCTK Data Download" button on the ExoCTK website.'
            )

        # If the variable exists but doesn't point to a real location
        if not os.path.exists(env_variables['exoctk_data']):
            raise FileNotFoundError(
                'The $EXOCTK_DATA environment variable is set to a location that '
                'cannot be accessed.')

        # If the variable exists, points to a real location, but is missing contents
        for item in ['modelgrid', 'fortney', 'exoctk_log', 'generic']:
            if item not in [os.path.basename(item) for item in glob.glob(os.path.join(env_variables['exoctk_data'], '*'))]:
                raise KeyError('Missing {}/ directory from {}'.format(item, env_variables['exoctk_data']))

    env_variables['modelgrid_dir'] = os.path.join(env_variables['exoctk_data'], 'modelgrid/')
    env_variables['fortgrid_dir'] = os.path.join(env_variables['exoctk_data'], 'fortney/')
    env_variables['exoctklog_dir'] = os.path.join(env_variables['exoctk_data'], 'exoctk_log/')
    env_variables['genericgrid_dir'] = os.path.join(env_variables['exoctk_data'], 'generic/')

    return env_variables


def get_target_data(target_name):
    """
    Send request to exomast restful api for target information.

    Parameters
    ----------
    target_name : string
        The name of the target transit

    Returns
    -------
    target_data: json:
        json object with target data.
    """

    canonical_name = get_canonical_name(target_name)

    target_url = build_target_url(canonical_name)

    r = requests.get(target_url)

    if r.status_code == 200:
        target_data = r.json()
    else:
        print('Whoops, no data for this target!')

    # Some targets have multiple catalogs
    # nexsci is the first choice.
    if len(target_data) > 1:
        # Get catalog names from exomast and make then the keys of a dictionary
        # and the values are its position in the json object.
        catalog_dict = {data['catalog_name']: index for index, data in enumerate(target_data)}

        # Parse based on catalog accuracy.
        if 'nexsci' in list(catalog_dict.keys()):
            target_data = target_data[catalog_dict['nexsci']]
        elif 'exoplanets.org' in list(catalog_dict.keys()):
            target_data = target_data[catalog_dict['exoplanets.org']]
        else:
            target_data = target_data[0]
    else:
        target_data = target_data[0]

    # Strip spaces and non numeric or alphabetic characters and combine.
    url = 'https://exo.mast.stsci.edu/exomast_planet.html?planet={}'.format(re.sub(r'\W+', '', canonical_name))

    return target_data, url



from functools import wraps
import os
import json
from pkg_resources import resource_filename

from astropy.coordinates import SkyCoord
from astropy.extern.six.moves import StringIO
import astropy.table as at
from astropy.time import Time
import astropy.units as u
from bokeh.resources import INLINE
from bokeh.util.string import encode_utf8
from bokeh.embed import components
import flask
from flask import Flask, Response, send_from_directory
from flask import request, send_file, make_response, render_template
import form_validation as fv
import numpy as np

from exoctk.contam_visibility import visibilityPA as vpa
from exoctk.contam_visibility import field_simulator as fs
from exoctk.contam_visibility import contamination_figure as cf
from exoctk.contam_visibility.miniTools import contamVerify
from exoctk.forward_models.forward_models import fortney_grid, generic_grid
from exoctk.groups_integrations.groups_integrations import perform_calculation
from exoctk.limb_darkening import limb_darkening_fit as lf
from exoctk.utils import filter_table, get_env_variables, get_target_data, get_canonical_name
from exoctk.modelgrid import ModelGrid
from exoctk.phase_constraint_overlap.phase_constraint_overlap import phase_overlap_constraint, calculate_pre_duration

from matplotlib.backends.backend_pdf import PdfPages

import log_exoctk
from svo_filters import svo

# FLASK SET UP
app_exoctk = Flask(__name__)

# define the cache config keys, remember that it can be done in a settings file
app_exoctk.config['CACHE_TYPE'] = 'null'
app_exoctk.config['SECRET_KEY'] = 'Thisisasecret!'


# Load the database to log all form submissions
if get_env_variables()['exoctklog_dir'] is None:
    dbpath = ':memory:'
else:
    dbpath = os.path.realpath(os.path.join(get_env_variables()['exoctklog_dir'], 'exoctk_log.db'))
    if not os.path.isfile(dbpath):
        log_exoctk.create_db(dbpath)
try:
    DB = log_exoctk.load_db(dbpath)
except IOError:
    DB = None


# Redirect to the index
@app_exoctk.route('/')
@app_exoctk.route('/index')
def index():
    """The Index page"""
    return render_template('index.html')


@app_exoctk.route('/limb_darkening', methods=['GET', 'POST'])
def limb_darkening():
    """The limb darkening form page. """
    # Load default form
    form = fv.LimbDarkeningForm()

    # Reload page with stellar data from ExoMAST
    if form.resolve_submit.data:

        if form.targname.data.strip() != '':

            try:

                # Resolve the target in exoMAST
                form.targname.data = get_canonical_name(form.targname.data)
                data, target_url = get_target_data(form.targname.data)

                # Update the form data
                form.feh.data = data.get('Fe/H')
                form.teff.data = data.get('Teff')
                form.logg.data = data.get('stellar_gravity')
                form.target_url.data = str(target_url)

            except Exception:
                form.target_url.data = ''
                form.targname.errors = ["Sorry, could not resolve '{}' in exoMAST.".format(form.targname.data)]

        # Send it back to the main page
        return render_template('limb_darkening.html', form=form)

    # Reload page with appropriate filter data
    if form.filter_submit.data:

        kwargs = {}
        if form.bandpass.data == 'tophat':
            kwargs['n_bins'] = 1
            kwargs['pixels_per_bin'] = 100
            kwargs['wave_min'] = 1 * u.um
            kwargs['wave_max'] = 2 * u.um

        # Get the filter
        bandpass = svo.Filter(form.bandpass.data, **kwargs)

        # Update the form data
        form.wave_min.data = bandpass.wave_min.value
        form.wave_max.data = bandpass.wave_max.value

        # Send it back to the main page
        return render_template('limb_darkening.html', form=form)

    # Update validation values after a model grid is selected
    if form.modelgrid_submit.data:

        # Load the modelgrid
        mg = ModelGrid(form.modeldir.data, resolution=500)
        teff_rng = mg.Teff_vals.min(), mg.Teff_vals.max()
        logg_rng = mg.logg_vals.min(), mg.logg_vals.max()
        feh_rng = mg.FeH_vals.min(), mg.FeH_vals.max()

        # Update the validation parameters by setting validator attributes
        setattr(form.teff.validators[1], 'min', float(teff_rng[0]))
        setattr(form.teff.validators[1], 'max', float(teff_rng[1]))
        setattr(form.teff.validators[1], 'message', 'Effective temperature must be between {} and {}'.format(*teff_rng))
        setattr(form.logg.validators[1], 'min', float(logg_rng[0]))
        setattr(form.logg.validators[1], 'max', float(logg_rng[1]))
        setattr(form.logg.validators[1], 'message', 'Surface gravity must be between {} and {}'.format(*logg_rng))
        setattr(form.feh.validators[1], 'min', float(feh_rng[0]))
        setattr(form.feh.validators[1], 'max', float(feh_rng[1]))
        setattr(form.feh.validators[1], 'message', 'Metallicity must be between {} and {}'.format(*feh_rng))

        # Send it back to the main page
        return render_template('limb_darkening.html', form=form)

    # Validate form and submit for results
    if form.validate_on_submit() and form.calculate_submit.data:

        # Get the stellar parameters
        star_params = [float(form.teff.data), float(form.logg.data), float(form.feh.data)]

        # Log the form inputs
        try:
            log_exoctk.log_form_input(request.form, 'limb_darkening', DB)
        except Exception:
            pass

        # Load the model grid
        model_grid = ModelGrid(form.modeldir.data, resolution=500)
        form.modeldir.data = [j for i, j in form.modeldir.choices if i == form.modeldir.data][0]

        # Grism details
        kwargs = {'n_bins': form.n_bins.data, 'wave_min': form.wave_min.data * u.um, 'wave_max': form.wave_max.data * u.um}

        # Make filter object and plot
        bandpass = svo.Filter(form.bandpass.data, **kwargs)
        # bp_name = bandpass.name
        bk_plot = bandpass.plot(draw=False)
        bk_plot.plot_width = 580
        bk_plot.plot_height = 280
        js_resources = INLINE.render_js()
        css_resources = INLINE.render_css()
        filt_script, filt_plot = components(bk_plot)

        # Trim the grid to nearby grid points to speed up calculation
        # full_rng = [model_grid.Teff_vals, model_grid.logg_vals, model_grid.FeH_vals]
        # trim_rng = find_closest(full_rng, star_params, n=1, values=True)

        # Calculate the coefficients for each profile
        ld = lf.LDC(model_grid)
        for prof in form.profiles.data:
            ld.calculate(*star_params, prof, mu_min=float(form.mu_min.data), bandpass=bandpass)

        # Draw tabbed figure
        final = ld.plot_tabs()

        # Get HTML
        script, div = components(final)

        # Store the tables as a string
        keep_cols = ['Teff', 'logg', 'FeH', 'profile', 'filter', 'wave_min', 'wave_eff', 'wave_max', 'c1', 'e1', 'c2', 'e2', 'c3', 'e3', 'c4', 'e4']
        print_table = ld.results[[col for col in keep_cols if col in ld.results.colnames]]
        file_as_string = '\n'.join(print_table.pformat(max_lines=-1, max_width=-1))

        # Make a table for each profile with a row for each wavelength bin
        profile_tables = []
        for profile in form.profiles.data:

            # Make LaTeX for polynomials
            latex = lf.ld_profile(profile, latex=True)
            poly = '\({}\)'.format(latex).replace('*', '\cdot').replace('\e', 'e')

            # Make the table into LaTeX
            table = filter_table(ld.results, profile=profile)
            co_cols = [c for c in ld.results.colnames if (c.startswith('c') or c.startswith('e')) and len(c) == 2 and not np.all([np.isnan(i) for i in table[c]])]
            table = table[['wave_eff', 'wave_min', 'wave_max'] + co_cols]
            table.rename_column('wave_eff', '\(\lambda_\mbox{eff}\hspace{5px}(\mu m)\)')
            table.rename_column('wave_min', '\(\lambda_\mbox{min}\hspace{5px}(\mu m)\)')
            table.rename_column('wave_max', '\(\lambda_\mbox{max}\hspace{5px}(\mu m)\)')

            # Add the results to the lists
            html_table = '\n'.join(table.pformat(max_width=-1, max_lines=-1, html=True)).replace('<table', '<table id="myTable" class="table table-striped table-hover"')

            # Add the table title
            header = '<br></br><strong>{}</strong><br><p>\(I(\mu)/I(\mu=1)\) = {}</p>'.format(profile, poly)
            html_table = header + html_table

            profile_tables.append(html_table)

        return render_template('limb_darkening_results.html', form=form,
                               table=profile_tables, script=script, plot=div,
                               file_as_string=repr(file_as_string),
                               filt_plot=filt_plot, filt_script=filt_script,
                               js=js_resources, css=css_resources)

    return render_template('limb_darkening.html', form=form)


@app_exoctk.route('/limb_darkening_error', methods=['GET', 'POST'])
def limb_darkening_error():
    """The limb darkening error page"""

    return render_template('limb_darkening_error.html')


@app_exoctk.route('/groups_integrations', methods=['GET', 'POST'])
def groups_integrations():
    """The groups and integrations calculator form page"""

    # Print out pandeia sat values
    with open(resource_filename('exoctk', 'data/groups_integrations/groups_integrations_input_data.json')) as f:
        sat_data = json.load(f)['fullwell']

    # Load default form
    form = fv.GroupsIntsForm()

    if request.method == 'GET':

        # http://0.0.0.0:5000/groups_integrations?k_mag=8.131&transit_duration=0.09089&target=WASP-18+b
        target_name = request.args.get('target')
        form.targname.data = target_name

        k_mag = request.args.get('k_mag')
        form.kmag.data = k_mag

        # According to Kevin the obs_dur = 3*trans_dur+1 hours
        # transit_dur is in days from exomast, convert first.
        try:
            trans_dur = float(request.args.get('transit_duration'))
            trans_dur *= u.day.to(u.hour)
            obs_dur = 3 * trans_dur + 1
            form.obs_duration.data = obs_dur
        except TypeError:
            trans_dur = request.args.get('transit_duration')
            if trans_dur is None:
                pass
            else:
                err = 'The Transit Duration from ExoMAST experienced some issues. Try a different spelling or source.'
                return render_template('groups_integrations_error.html', err=err)
        return render_template('groups_integrations.html', form=form, sat_data=sat_data)

    # Reload page with stellar data from ExoMAST
    if form.resolve_submit.data:

        if form.targname.data.strip() != '':

            # Resolve the target in exoMAST
            try:
                form.targname.data = get_canonical_name(form.targname.data)
                data, url = get_target_data(form.targname.data)

                # Update the Kmag
                kmag = data.get('Kmag')

                # Transit duration in exomast is in days, need it in hours
                if form.time_unit.data == 'day':
                    trans_dur = data.get('transit_duration')
                    obs_dur = 3 * trans_dur + (1 / 24.)
                else:
                    trans_dur = data.get('transit_duration')
                    trans_dur *= u.Unit('day').to('hour')
                    obs_dur = 3 * trans_dur + 1

                # Model guess
                logg_targ = data.get('stellar_gravity') or 4.5
                teff_targ = data.get('Teff') or 5500
                arr = np.array([tuple(i[1].split()) for i in form.mod.choices], dtype=[('spt', 'O'), ('teff', '>f4'), ('logg', '>f4')])
                mod_table = at.Table(arr)

                # If high logg, remove giants from guess list
                if logg_targ < 4:
                    mod_table = filter_table(mod_table, logg=">=4")
                teff = min(arr['teff'], key=lambda x: abs(x - teff_targ))
                mod_table.add_column(at.Column(np.array([i[0] for i in form.mod.choices]), name='value'))
                mod_table = filter_table(mod_table, teff="<={}".format(teff))
                mod_table.sort(['teff', 'logg'])

                # Set the form values
                form.mod.data = mod_table[-1]['value']
                form.kmag.data = kmag
                form.obs_duration.data = obs_dur
                form.target_url.data = url

            except Exception:
                form.target_url.data = ''
                form.targname.errors = ["Sorry, could not resolve '{}' in exoMAST.".format(form.targname.data)]

        # Send it back to the main page
        return render_template('groups_integrations.html', form=form, sat_data=sat_data)

    if form.validate_on_submit() and form.calculate_submit.data:

        # Get the form data
        ins = form.ins.data
        params = {'ins': ins,
                  'mag': form.kmag.data,
                  'obs_time': form.obs_duration.data,
                  'sat_max': form.sat_max.data,
                  'sat_mode': form.sat_mode.data,
                  'time_unit': form.time_unit.data,
                  'band': 'K',
                  'mod': form.mod.data,
                  'filt'.format(ins): getattr(form, '{}_filt'.format(ins)).data,
                  'subarray'.format(ins): getattr(form, '{}_subarray'.format(ins)).data,
                  'filt_ta'.format(ins): getattr(form, '{}_filt_ta'.format(ins)).data,
                  'subarray_ta'.format(ins): getattr(form, '{}_subarray_ta'.format(ins)).data}

        # Get ngroups
        params['n_group'] = 'optimize' if form.n_group.data == 0 else int(form.n_group.data)

        # Also get the data path in there
        params['infile'] = resource_filename('exoctk', 'data/groups_integrations/groups_integrations_input_data.json')

        # Convert the obs_time to hours
        if params['time_unit'] == 'day':
            params['obs_time'] = params['obs_time'] * 24
            params['time_unit'] = 'hours'

        # Run the calculation
        results = perform_calculation(params)
        if type(results) == dict:
            results_dict = results
            one_group_error = ""
            zero_group_error = ""
            if results_dict['n_group'] == 1:
                one_group_error = 'Be careful! This only predicts one group, and you may be in danger of oversaturating!'
            if results_dict['max_ta_groups'] == 0:
                zero_group_error = 'Be careful! This oversaturated the TA in the minimum groups. Consider a different TA setup.'
            if results_dict['max_ta_groups'] == -1:
                zero_group_error = 'This object is too faint to reach the required TA SNR in this filter. Consider a different TA setup.'
                results_dict['min_sat_ta'] = 0
                results_dict['t_duration_ta_max'] = 0
                results_dict['max_sat_ta'] = 0
                results_dict['t_duration_ta_max'] = 0
            if results_dict['max_sat_prediction'] > results_dict['sat_max']:
                one_group_error = 'This many groups will oversaturate the detector! Proceed with caution!'
            # Do some formatting for a prettier end product
            results_dict['filt'] = results_dict['filt'].upper()
            results_dict['filt_ta'] = results_dict['filt_ta'].upper()
            results_dict['band'] = results_dict['band'].upper()
            results_dict['mod'] = results_dict['mod'].upper()
            if results_dict['ins'] == 'niriss':
                if results_dict['subarray_ta'] == 'nrm':
                    results_dict['subarray_ta'] = 'SUBTASOSS -- BRIGHT'
                else:
                    results_dict['subarray_ta'] = 'SUBTASOSS -- FAINT'
            results_dict['subarray'] = results_dict['subarray'].upper()
            results_dict['subarray_ta'] = results_dict['subarray_ta'].upper()

            form_dict = {'miri': 'MIRI', 'nircam': 'NIRCam', 'nirspec': 'NIRSpec', 'niriss': 'NIRISS'}
            results_dict['ins'] = form_dict[results_dict['ins']]

            return render_template('groups_integrations_results.html',
                                   results_dict=results_dict,
                                   one_group_error=one_group_error,
                                   zero_group_error=zero_group_error)

        else:
            err = results
            return render_template('groups_integrations_error.html', err=err)

    return render_template('groups_integrations.html', form=form, sat_data=sat_data)


@app_exoctk.route('/contam_visibility', methods=['GET', 'POST'])
def contam_visibility():
    """The contamination and visibility form page"""
    # Load default form
    form = fv.ContamVisForm()
    form.calculate_contam_submit.disabled = False

    if request.method == 'GET':

        # http://0.0.0.0:5000/contam_visibility?ra=24.354208334287005&dec=-45.677930555343636&target=WASP-18%20b
        target_name = request.args.get('target')
        form.targname.data = target_name

        ra = request.args.get('ra')
        form.ra.data = ra

        dec = request.args.get('dec')
        form.dec.data = dec

        return render_template('contam_visibility.html', form=form)

    # Reload page with stellar data from ExoMAST
    if form.resolve_submit.data:

        if form.targname.data.strip() != '':

            # Resolve the target in exoMAST
            try:
                form.targname.data = get_canonical_name(form.targname.data)
                data, url = get_target_data(form.targname.data)

                # Update the coordinates
                ra_deg = data.get('RA')
                dec_deg = data.get('DEC')

                # Set the form values
                form.ra.data = ra_deg
                form.dec.data = dec_deg
                form.target_url.data = url

            except Exception:
                form.target_url.data = ''
                form.targname.errors = ["Sorry, could not resolve '{}' in exoMAST.".format(form.targname.data)]

        # Send it back to the main page
        return render_template('contam_visibility.html', form=form)

    # Reload page with appropriate mode data
    if form.mode_submit.data:

        # Update the button
        if ('NIRCam' in form.inst.data) or (form.inst.data == 'MIRI') or (form.inst.data == 'NIRSpec'):
            form.calculate_contam_submit.disabled = True
        else:
            form.calculate_contam_submit.disabled = False

        # Send it back to the main page
        return render_template('contam_visibility.html', form=form)

    if form.validate_on_submit() and (form.calculate_submit.data or form.calculate_contam_submit.data):

        try:

            # Log the form inputs
            try:
                log_exoctk.log_form_input(request.form, 'contam_visibility', DB)
            except Exception:
                pass

            # Make plot
            title = form.targname.data or ', '.join([str(form.ra.data), str(form.dec.data)])
            pG, pB, dates, vis_plot, table, badPAs = vpa.using_gtvt(str(form.ra.data),
                                                                    str(form.dec.data),
                                                                    form.inst.data.split(' ')[0],
                                                                    targetName=str(title))

            # Make output table
            fh = StringIO()
            table.write(fh, format='csv', delimiter=',')
            visib_table = fh.getvalue()

            # Get scripts
            vis_js = INLINE.render_js()
            vis_css = INLINE.render_css()
            vis_script, vis_div = components(vis_plot)

            # Contamination plot too
            if form.calculate_contam_submit.data:

                # First convert ra and dec to HH:MM:SS
                ra_deg, dec_deg = float(form.ra.data), float(form.dec.data)
                sc = SkyCoord(ra_deg, dec_deg, unit='deg')
                ra_dec = sc.to_string('hmsdms')
                ra_hms, dec_dms = ra_dec.split(' ')[0], ra_dec.split(' ')[1]

                # Make field simulation
                contam_cube = fs.fieldSim(ra_hms, dec_dms, form.inst.data, binComp=form.companion.data)
                contam_plot = cf.contam(contam_cube, form.inst.data, targetName=str(title), paRange=[int(form.pa_min.data), int(form.pa_max.data)], badPAs=badPAs, fig='bokeh')

                # Get scripts
                contam_js = INLINE.render_js()
                contam_css = INLINE.render_css()
                contam_script, contam_div = components(contam_plot)

            else:

                contam_script = contam_div = contam_js = contam_css = ''

            return render_template('contam_visibility_results.html',
                                   form=form, vis_plot=vis_div,
                                   vis_table=visib_table,
                                   vis_script=vis_script, vis_js=vis_js,
                                   vis_css=vis_css, contam_plot=contam_div,
                                   contam_script=contam_script,
                                   contam_js=contam_js,
                                   contam_css=contam_css)

        except Exception as e:
            err = 'The following error occurred: ' + str(e)
            return render_template('groups_integrations_error.html', err=err)

    return render_template('contam_visibility.html', form=form)


@app_exoctk.route('/visib_result', methods=['POST'])
def save_visib_result():
    """Save the results of the Visibility Only calculation"""

    visib_table = flask.request.form['data_file']
    targname = flask.request.form['targetname']
    targname = targname.replace(' ', '_') # no spaces
    instname = flask.request.form['instrumentname']

    return flask.Response(visib_table, mimetype="text/dat",
                          headers={"Content-disposition": "attachment; filename={}_{}_visibility.csv".format(targname, instname)})

@app_exoctk.route('/contam_verify', methods=['GET', 'POST'])
def save_contam_pdf():
    """Save the results of the Contamination Science FOV """

    RA, DEC = '19:50:50.2400', '+48:04:51.00'
    contam_pdf = contamVerify(RA, DEC, 'NIRISS', [1,2], binComp=[], PDF='', web=True)

    filename = contam_pdf.split('/')[-1]
    pdf_obj = PdfPages(contam_pdf)

    return render_template(contam_pdf, filename, as_attachment=True)#, mimetype="application/pdf", as_attachment=True)
    #return flask.Response(pdf_obj, mimetype="application/pdf",
    #                      headers={"Content-disposition": "attachment; filename={}_{}_contam.pdf".format(targname, instname)})


@app_exoctk.route('/download', methods=['POST'])
def exoctk_savefile():
    """Save results to file"""

    file_as_string = eval(request.form['file_as_string'])

    response = make_response(file_as_string)
    response.headers["Content-type"] = 'text; charset=utf-8'
    response.headers["Content-Disposition"] = "attachment; filename=ExoCTK_results.txt"
    return response


def _param_fort_validation(args):
    """Validates the input parameters for the forward models"""

    temp = args.get('ptemp', 1000)
    chem = args.get('pchem', 'noTiO')
    cloud = args.get('cloud', '0')
    pmass = args.get('pmass', '1.5')
    m_unit = args.get('m_unit', 'M_jup')
    reference_radius = args.get('refrad', 1)
    r_unit = args.get('r_unit', 'R_jup')
    rstar = args.get('rstar', 1)
    rstar_unit = args.get('rstar_unit', 'R_sun')

    input_args = {'temp': temp, 'chem': chem, 'cloud': cloud, 'pmass': pmass,
                  'm_unit': m_unit, 'reference_radius': reference_radius,
                  'r_unit': r_unit, 'rstar': rstar, 'rstar_unit': rstar_unit}

    return input_args


@app_exoctk.route('/fortney', methods=['GET', 'POST'])
def fortney():
    """
    Pull up Forntey Grid plot the results and download
    """

    # Grab the inputs arguments from the URL
    args = flask.request.args

    input_args = _param_fort_validation(args)
    fig, fh, temp_out = fortney_grid(input_args)

    table_string = fh.getvalue()

    js_resources = INLINE.render_js()
    css_resources = INLINE.render_css()

    script, div = components(fig)

    html = flask.render_template('fortney.html',
                                 plot_script=script,
                                 plot_div=div,
                                 js_resources=js_resources,
                                 css_resources=css_resources,
                                 temp=temp_out,
                                 table_string=table_string
                                 )
    return encode_utf8(html)


@app_exoctk.route('/generic', methods=['GET', 'POST'])
def generic():
    """
    Pull up Generic Grid plot the results and download
    """

    # Grab the inputs arguments from the URL
    args = dict(flask.request.args)
    fig, fh, closest_match, error_message = generic_grid(args)

    # Write table string
    table_string = fh.getvalue()

    # Web-ify bokeh plot
    js_resources = INLINE.render_js()
    css_resources = INLINE.render_css()

    script, div = components(fig)

    html = flask.render_template('generic.html',
                                 inputs=args,
                                 closest_match=closest_match,
                                 error_message=error_message,
                                 table_string=table_string,
                                 plot_script=script,
                                 plot_div=div,
                                 js_resources=js_resources,
                                 css_resources=css_resources,
                                 )
    return encode_utf8(html)


@app_exoctk.route('/fortney_result', methods=['POST'])
def save_fortney_result():
    """SAve the results of the Fortney grid"""

    table_string = flask.request.form['data_file']
    return flask.Response(table_string, mimetype="text/dat", headers={"Content-disposition": "attachment; filename=fortney.dat"})


@app_exoctk.route('/generic_result', methods=['POST'])
def save_generic_result():
    """Save the results of the generic grid"""

    table_string = flask.request.form['data_file']
    return flask.Response(table_string, mimetype="text/dat", headers={"Content-disposition": "attachment; filename=generic.dat"})


@app_exoctk.route('/groups_integrations_download')
def groups_integrations_download():
    """Download the groups and integrations calculator data"""

    return send_file(resource_filename('exoctk', 'data/groups_integrations/groups_integrations_input_data.json'), mimetype="text/json", attachment_filename='groups_integrations_input_data.json', as_attachment=True)


@app_exoctk.route('/fortney_download')
def fortney_download():
    """Download the fortney grid data"""

    fortney_data = os.path.join(get_env_variables()['fortgrid_dir'], 'fortney_grid.db')
    return send_file(fortney_data, attachment_filename='fortney_grid.db', as_attachment=True)


def check_auth(username, password):
    """This function is called to check if a username password combination is
    valid

    Parameters
    ----------
    username: str
        The username
    password: str
        The password
    """

    return username == 'admin' and password == 'secret'


def authenticate():
    """Sends a 401 response that enables basic auth"""

    return Response('Could not verify your access level for that URL.\n'
                    'You have to login with proper credentials', 401,
                    {'WWW-Authenticate': 'Basic realm="Login Required"'})


def requires_auth(page):
    """Requires authentication for a page before loading

    Parameters
    ----------
    page: function
        The function that sets a route

    Returns
    -------
    function
        The decorated route
    """

    @wraps(page)
    def decorated(*args, **kwargs):
        auth = request.authorization
        if not auth or not check_auth(auth.username, auth.password):
            return authenticate()
        return page(*args, **kwargs)
    return decorated


@app_exoctk.route('/admin')
@requires_auth
def secret_page():
    """Shhhhh! This is a secret page of admin stuff"""

    tables = [i[0] for i in DB.execute("SELECT name FROM sqlite_master WHERE type='table'").fetchall()]

    log_tables = []
    for table in tables:

        try:
            data = log_exoctk.view_log(DB, table)

            # Add the results to the lists
            html_table = '\n'.join(data.pformat(max_width=500, html=True)).replace('<table', '<table id="myTable" class="table table-striped table-hover"')

        except Exception:
            html_table = '<p>No data to display</p>'

        # Add the table title
        header = '<h3>{}</h3>'.format(table)
        html_table = header + html_table

        log_tables.append(html_table)

    return render_template('admin_page.html', tables=log_tables)


@app_exoctk.route('/lightcurve_fitting')
def lightcurve_fitting():
    """A landing page for the lightcurve_fitting tool"""

    return render_template('lightcurve_fitting.html')


@app_exoctk.route('/atmospheric_retrievals')
def atmospheric_retrievals():
    """A landing page for the atmospheric_retrievals tools"""

    return render_template('atmospheric_retrievals.html')


@app_exoctk.route('/phase_constraint', methods=['GET', 'POST'])
def phase_constraint(transit_type='primary'):
    # Load default form
    form = fv.PhaseConstraint()

    # Reload page with stellar data from ExoMAST
    if form.resolve_submit.data:
        if form.targname.data.strip() != '':
            try:
                # Resolve the target in exoMAST
                form.targname.data = get_canonical_name(form.targname.data)
                data, target_url = get_target_data(form.targname.data)

                # Update the form data
                form.orbital_period.data = data.get('orbital_period')

                t_time = Time(data.get('transit_time'), format='mjd')
                form.transit_time.data = t_time.jd

                form.observation_duration.data = calculate_pre_duration(data.get('transit_duration') * 24.0)

                form.inclination.data = data.get('inclination')
                if form.inclination.data is None:
                    form.inclination.data = np.nan

                form.omega.data = data.get('omega')
                if form.omega.data is None:
                    form.omega.data = np.nan

                form.eccentricity.data = data.get('eccentricity')
                if form.eccentricity.data is None:
                    form.omega.data = np.nan

                form.target_url.data = str(target_url)

                return render_template('phase_constraint.html', form=form)

            except Exception:
                form.target_url.data = ''
                form.targname.errors = ["Sorry, could not resolve '{}' in exoMAST.".format(form.targname.data)]

    # Extract transit type:
    transit_type = form.transit_type.data
    if form.validate_on_submit() and form.calculate_submit.data:
        if transit_type == 'primary':
            minphase, maxphase = phase_overlap_constraint(target_name=form.targname.data,
                                                          period=form.orbital_period.data,
                                                          pretransit_duration=form.observation_duration.data,
                                                          window_size=form.window_size.data)
        elif transit_type == 'secondary':
            if (0. <= form.eccentricity.data < 1) and (-360. <= form.omega.data <= 360.) and (0 <= form.inclination.data <= 90.):
                # Use dummy time-of-transit as it doesn't matter for the phase-constraint calculation
                # (phase = 1 is always transit)
                minphase, maxphase = phase_overlap_constraint(target_name=form.targname.data,
                                                              period=form.orbital_period.data, t0=1.,
                                                              pretransit_duration=form.observation_duration.data,
                                                              window_size=form.window_size.data, secondary=True,
                                                              ecc=form.eccentricity.data, omega=form.omega.data,
                                                              inc=form.inclination.data)
            else:
                minphase, maxphase = np.nan, np.nan
        else:
            minphase, maxphase = np.nan, np.nan
        form.minimum_phase.data = minphase
        form.maximum_phase.data = maxphase
    """
    if (form.eccentricity.data > 1.) or (form.eccentricity.data < 0.):
        form.eccentricity.data = None
    if np.abs(form.omega.data)>360.:
        form.omega.data = None
    if (form.inclination.data < 0) or (form.inclination.data>90.):
        form.inclination.data = None
    """
    # Send it back to the main page
    return render_template('phase_constraint.html', form=form)


if __name__ == '__main__':
    # os.chmod('/internal/data1/app_data/.astropy/cache/', 777)
    port = int(os.environ.get('PORT', 5000))
    app_exoctk.run(host='0.0.0.0', port=port)



# ! /usr/bin/env python
# quaternion module
"""Version 4 September 9, 2010 WMK
Flipped sign of the angle in the QX, QY, QZ, QJX, QJY, QJZ, set_values,
set_as_QX, ... functions to be consistent with the corrected multiplication.
Also updated
the doc strings.

Version 3 September 8, 2010 RLH
Backed out change to cnvrt in version 2.

Version 2 September 3, 2010 RLH
Fixed sign error in quaternion multiplication functions.
Added quaternion __str__ method.
Modified cnvrt method to return a CelestialVector


4/9/2010 WMK
Redefined the __init__ inputs
Changed from a Vector internal representation to 3 scalers
Fixed an error in cvt_att_Q_to_angles, was assuming an att2inertial Quaternion!
Streamlined some of the functions
 Version 1.0 August 3, 2010
 Got rid of degrees trig functions.

Combined this and rotationsx.py module to avoid circular imports and made it
PEP Compliant
Joe Filippazzo - 2018/06/26
"""
from math import radians, asin, cos, sin, sqrt, pi, degrees, atan2

from . import math_extensionsx as math2
from .astro_funcx import unit_limit

D2R = pi / 180.
R2D = 180. / pi
PI2 = 2. * pi


class GalacticPole:
    """Represents coordinates of galactic pole."""

    def __init__(self, latitude, longitude, ascending_node):
        """Initializes the coordinates of the galactic pole.

        Parameters
        ----------
        latitude: float
            Latitude of pole, in degrees.
        longitude: float
            Longitude of pole, in degrees.
        ascending_node: float
            Ascending node of pole, in degrees.
        """
        # Arguments specified in degrees, but values represented in radians.
        self.latitude = radians(latitude)
        self.longitude = radians(longitude)
        self.anode = radians(ascending_node)

    def __str__(self):
        """Returns string representation of the galactic pole."""
        text = (degrees(self.latitude), degrees(self.longitude),
                degrees(self.anode))
        # Convert attributes back into degrees for readability.
        return """GalacticPole: latitude: %.3fD, longitude: %.3fD,\
                  anode: %.3fD""" % text


# supports transformation to galactic coordinates
NGP = GalacticPole(192.859508, 27.128336, 32.932)


def QX(angle):
    """Creates rotation quaternion about X axis, rotates a vector about
    this axis.

    Parameters
    ----------
    angle: float
        The angle to rotate by.

    Result
    ------
    Quarternion
        The rotated quaternion.
    """
    return Quaternion(Vector(sin(angle / 2.), 0., 0.), cos(angle / 2.))


def QY(angle):
    """Creates rotation quaternion about Y axis, rotates a vector about
    this axis.

    Parameters
    ----------
    angle: float
        The angle to rotate by.

    Result
    ------
    Quarternion
        The rotated quaternion.
    """
    return Quaternion(Vector(0., sin(angle / 2.), 0.), cos(angle / 2.))


def QZ(angle):
    """Creates rotation quaternion about Z axis, rotates a vector about
    this axis

    Parameters
    ----------
    angle: float
        The angle to rotate by.

    Result
    ------
    Quarternion
        The rotated quaternion.
    """
    return Quaternion(Vector(0., 0., sin(angle / 2.)), cos(angle / 2.))


def Qmake_a_point(V):
    """Creates a pure Q, i.e. defines a pointing not a rotation

    Parameters
    ----------
    V: Vector
        The vector.

    Returns
    -------
    Quaternion
        The point as a quaternion.
    """
    return Quaternion(V, 0.)


def cvt_pt_Q_to_V(Q):
    """Converts a pure (pointing) Q to a unit position Vector

    Parameters
    ----------
    Q: Quaternion
        The quaternion to convert to a Vector.

    Returns
    -------
    Vector
        The point as a vector.
    """
    return Vector(Q.q1, Q.q2, Q.q3)


# The following functions are dependent upon the spacecraft definitions
# and perhaps should be moved to that module
def Qmake_body2inertial(coord1, coord2, V3pa):
    """Creates a rotation Q, going from the body frame to inertial.

    Parameters
    ----------
    coord1: float
        The first coordinate.
    coord2: float
        The second coordinate.
    V3pa: float
        The V3 position.

    Returns
    -------
    Quaternion
        The rotation quaternion
    """
    return QZ(coord1) * QY(-coord2) * QX(-V3pa)


def Qmake_v2v3_2body(v2, v3):
    """Creates a rotation Q, going from v2 and v3 in the body frame to
    inertial.

    Parameters
    ----------
    v2: float
        The V2 position.
    V3: float
        The V3 position.

    Returns
    -------
    Quaternion
        The rotation quaternion.
    """
    return QY(v3) * QZ(-v2)


def Qmake_v2v3_2inertial(coord1, coord2, V3pa, v2, v3):
    """Creates a rotation Q, going from v2 and v3 in the body frame to
    inertial

    Parameters
    ----------
    coord1: float
        The first coordinate.
    coord2: float
        The second coordinate.
    V3pa: float
        The V3 position.
    v2: float
        The V2 position.
    v3: float
        The V3 position.

    Returns
    -------
    Quaternion
        The rotation quaternion.
    """
    return QZ(coord1) * QY(-coord2) * QX(-V3pa) * QY(v3) * QZ(-v2)


def Qmake_aperture2inertial(coord1, coord2, APA, xoff, yoff, s, YapPA,
                            V3ref, V2ref):
    """Creates a rotation Q, going from the target in aperture frame to body.

    Parameters
    ----------
    coord1: float
        The first coordinate.
    coord2: float
        The second coordinate.
    APA: float
        The apature position.
    xoff: float
        The x offset.
    yoff: float
        The y offset.
    s: float
        The multiplicative factor.
    V2ref: float
        The V2 position.
    V3ref: float
        The V3 position.

    Returns
    -------
    Quaternion
        The rotation quaternion.
    """
    term1 = QZ(coord1) * QY(-coord2) * QX(-APA) * QY(-yoff)
    term2 = QZ(s * xoff) * QX(YapPA) * QY(V3ref) * QZ(-V2ref)
    return term1 * term2


def cvt_body2inertial_Q_to_c1c2pa_tuple(Q):
    """Creates a angle tuple from Q, assuming body frame to inertial Q and
    321 rotation sequence.

    Parameters
    ----------
    Q: Quaternion
        The quaternion.

    Returns
    -------
    coord1 : float
        The first coordinate.
    coord2 : float
        The second coordinate.
    pa : float
        The poosition angle.
    """
    # Conversion from Euler symmetric parameters to matrix elements and
    # matrix elements to rotation angles is given in Isaac's papers
    r11 = Q.q1 * Q.q1 - Q.q2 * Q.q2 - Q.q3 * Q.q3 + Q.q4 * Q.q4
    r21 = 2. * (Q.q1 * Q.q2 + Q.q3 * Q.q4)
    r31 = 2. * (Q.q1 * Q.q3 - Q.q2 * Q.q4)
    r32 = 2. * (Q.q2 * Q.q3 + Q.q1 * Q.q4)
    r33 = -Q.q1 * Q.q1 - Q.q2 * Q.q2 + Q.q3 * Q.q3 + Q.q4 * Q.q4
    coord1 = atan2(r21, r11)
    if coord1 < 0.:
        coord1 += PI2
    coord2 = math2.asin2(r31)  # use "safe" version of sine
    pa = atan2(-r32, r33)
    if pa < 0.:
        pa += PI2
    return coord1, coord2, pa


def cvt_v2v3_using_body2inertial_Q_to_c1c2pa_tuple(Q, v2, v3):
    """Given Q and v2, v3 gives pos on sky and V3 PA.

    Parameters
    ----------
    Q: Quaternion
        The quaternion.
    v2: float
        The V2 position.
    v3: float
        The V3 position.

    Returns
    -------
    tuple
        The coordinates and position angle
    """
    Vp_body = Vector(0., 0., 0.)
    Vp_body.set_xyz_from_angs(v2, v3)
    Vp_eci_pt = Q.cnvrt(Vp_body)
    coord1 = atan2(Vp_eci_pt.y, Vp_eci_pt.x)
    if coord1 < 0.:
        coord1 += PI2
    coord2 = asin(unit_limit(Vp_eci_pt.z))

    V3_body = Vector(0., 0., 1.)
    V3_eci_pt = Q.cnvrt(V3_body)
    NP_eci = Vector(0., 0., 1.)
    V_left = cross(NP_eci, Vp_eci_pt)
    if V_left.length() > 0.:
        V_left = V_left / V_left.length()
    NP_in_plane = cross(Vp_eci_pt, V_left)
    x = dot(V3_eci_pt, NP_in_plane)
    y = dot(V3_eci_pt, V_left)
    pa = atan2(y, x)
    if pa < 0.:
        pa += PI2

    return coord1, coord2, pa


def cvt_c1c2_using_body2inertial_Q_to_v2v3pa_tuple(Q, coord1, coord2):
    """Given Q and a position, returns v2, v3, V3PA tuple

    Parameters
    ----------
    Q: Quaternion
        The quaternion.
    coord1: float
        The first coordinate.
    coord2: float
        The second coordinate

    Returns
    -------
    coord1 : float
        The first coordinate.
    coord2 : float
        The second coordinate.
    pa : float
        The poosition angle.
    """
    Vp_eci = Vector(1., 0., 0.)
    Vp_eci.set_xyz_from_angs(coord1, coord2)
    Vp_body_pt = Q.inv_cnvrt(Vp_eci)
    v2 = atan2(Vp_body_pt.y, Vp_body_pt.x)
    v3 = asin(unit_limit(Vp_body_pt.z))
    V3_body = Vector(0., 0., 1.)
    V3_eci_pt = Q.cnvrt(V3_body)
    NP_eci = Vector(0., 0., 1.)
    V_left = cross(NP_eci, Vp_eci)
    if V_left.length() > 0.:
        V_left = V_left / V_left.length()
    NP_in_plane = cross(Vp_eci, V_left)
    x = dot(V3_eci_pt, NP_in_plane)
    y = dot(V3_eci_pt, V_left)
    pa = atan2(y, x)
    if pa < 0.:
        pa += PI2
    return v2, v3, pa


class Quaternion:
    """This representation is used by Wertz and Markley"""

    def __init__(self, V, q4):
        """Quaternion constructor.

        Parameters
        ----------
        V: Vector
            The vector to construct the quaternion with.
        q4: Vector
            The fourth vector.
        """
        self.q1 = V.x
        self.q2 = V.y
        self.q3 = V.z
        self.q4 = q4

    def __str__(self):
        """Returns a string representation of the quaternion."""
        text = (self.q1, self.q2, self.q3, self.q4)
        return 'Quaternion: q1: %.3f, q2: %.3f, q3: %.3f, q4: %.3f' % text

    def length(self):
        """Returns length of the Q """
        a = self.q1 * self.q1
        b = self.q2 * self.q2
        c = self.q3 * self.q3
        d = self.q4 * self.q4
        return sqrt(a + b + c + d)

    def normalize(self):
        """Returns a copy of the Q normalized """
        scale = self.length()
        return Quaternion(
            Vector(
                self.q1 / scale,
                self.q2 / scale,
                self.q3 / scale),
            self.q4 / scale)

    def conjugate(self):
        """Returns a copy of the conjugated Q """
        return Quaternion(Vector(-self.q1, -self.q2, -self.q3), self.q4)

    def __mul__(self, rs):
        """Defines Q*Q for quaternion multiplication.

        Parameters
        ----------
        rs: Quaternion
            The quaternion to multiply.

        Returns
        -------
        Q : Quaternion
            The multiplied Quaternion.
        """
        Q = Quaternion(Vector(0., 0., 0.), 0.)
        # Q.V = rs.V*self.q4 + self.V*rs.q4 + cross(self.V, rs.V)
        Q.q1 = rs.q1 * self.q4 + self.q1 * rs.q4 + \
            (self.q2 * rs.q3 - self.q3 * rs.q2)
        Q.q2 = rs.q2 * self.q4 + self.q2 * rs.q4 + \
            (self.q3 * rs.q1 - self.q1 * rs.q3)
        Q.q3 = rs.q3 * self.q4 + self.q3 * rs.q4 + \
            (self.q1 * rs.q2 - self.q2 * rs.q1)
        Q.q4 = self.q4 * rs.q4 - \
            (self.q1 * rs.q1 + self.q2 * rs.q2 + self.q3 * rs.q3)
        return Q

    def cnvrt(self, V):
        """Rotates a vector from the starting frame to the ending frame
        defined by the Q.

        Parameters
        ----------
        V: Vector
            The vector to rotate.

        Returns
        -------
        Vector
            The rotated Vector.
        """
        QV = Qmake_a_point(V)
        QV = self * QV * self.conjugate()
        return Vector(QV.q1, QV.q2, QV.q3)

    def inv_cnvrt(self, V):
        """Rotates a vector from the ending frame to the starting frame
        defined by the Q.

        Parameters
        ----------
        V: Vector
            The vector to invert.

        Returns
        -------
        Vector
            The inverted Vector.
        """
        QV = Qmake_a_point(V)
        QV = self.conjugate() * QV * self
        return Vector(QV.q1, QV.q2, QV.q3)

    def set_values(self, V, angle):
        """Sets quaterion values using a direction vector and a rotation of
        the coordinate frame about it.

        Parameters
        ----------
        V: Vector
            The direction Vector.
        angle: float
            The angle of rotation.
        """
        S = sin(-angle / 2.)
        self.q1 = V.x * S
        self.q2 = V.y * S
        self.q3 = V.z * S
        self.q4 = cos(angle / 2.)

    def set_as_QX(self, angle):
        """Sets quaterion in place like QX function.

        Parameters
        ----------
        angle: float
            The angle of rotation
        """
        self.q1 = sin(-angle / 2.)
        self.q2 = 0.
        self.q3 = 0.
        self.q4 = cos(angle / 2.)

    def set_as_QY(self, angle):
        """Sets quaterion in place like QY function

        Parameters
        ----------
        angle: float
            The angle of rotation
        """
        self.q1 = 0.
        self.q2 = sin(-angle / 2.)
        self.q3 = 0.
        self.q4 = cos(angle / 2.)

    def set_as_QZ(self, angle):
        """Sets quaterion in place like QZ function.

        Parameters
        ----------
        angle: float
            The angle of rotation.
        """
        self.q1 = 0.
        self.q2 = 0.
        self.q3 = sin(-angle / 2.)
        self.q4 = cos(angle / 2.)

    def set_as_mult(self, QQ1, QQ2):
        """Sets self as QQ1*QQ2 in place for quaternion multiplication.

        Parameters
        ----------
        QQ1: Quaternion
            The first quaternion.
        QQ2: Quaternion
            The second quaternion.
        """
        a = QQ1.q2 * QQ2.q3 - QQ1.q3 * QQ2.q2
        b = QQ1.q3 * QQ2.q1 - QQ1.q1 * QQ2.q3
        c = QQ1.q1 * QQ2.q2 - QQ1.q2 * QQ2.q1
        d = QQ1.q1 * QQ2.q1 + QQ1.q2 * QQ2.q2 + QQ1.q3 * QQ2.q3
        self.q1 = QQ2.q1 * QQ1.q4 + QQ1.q1 * QQ2.q4 + a
        self.q2 = QQ2.q2 * QQ1.q4 + QQ1.q2 * QQ2.q4 + b
        self.q3 = QQ2.q3 * QQ1.q4 + QQ1.q3 * QQ2.q4 + c
        self.q4 = QQ1.q4 * QQ2.q4 - d

    def set_as_point(self, V):
        """Set V as a point.

        Parameters
        ----------
        V: Vector
            The vector to set as a point.
        """
        self.q1 = V.x
        self.q2 = V.y
        self.q3 = V.z
        self.q4 = 0.

    def set_equal(self, Q):
        """Assigns values from other Q to this one.

        Parameters
        ----------
        Q: Quaternion
            The quaternion value to set.
        """
        self.q1 = Q.q1
        self.q2 = Q.q2
        self.q3 = Q.q3
        self.q4 = Q.q4

    def set_as_conjugate(self):
        """Assigns conjugate values in place. """
        self.q1 *= -1.
        self.q2 *= -1.
        self.q3 *= -1.


class NumericList(list):
    """List class that supports multiplication.  Only valid for numbers."""

    def __mul__(L1, L2):
        """Take the dot product of two numeric lists.
        Not using Vector for this because it is limited to three dimensions.
        Lists must have the same number of elements

        Parameters
        ----------
        L1: sequence
            The first list.
        L2: sequence
            The second list.

        Returns
        -------
        float
            The sum of the lists.
        """
        return(sum(map(lambda x, y: x * y, L1, L2)))


class Matrix(list):
    """Class to encapsulate matrix data and methods.

    A matrix is simply a list of lists that correspond to rows of the matrix.
    This is just intended to handle simple multiplication and vector rotations.
    For anything more advanced or computationally intensive, Python library
    routines should be used."""

    def __init__(self, rows):
        """Constructor for a matrix.

        This accepts a list of rows.
        It is assumed the rows are all of the same length.

        Parameters
        ----------
        rows: sequence
            The rows of the matrix.
        """
        for row in rows:
            self.append(NumericList(row))  # copy list

    def __str__(self):
        """Returns a string representation of the matrix."""
        return_str = 'Matrix:'

        for row_index in range(len(self)):
            row_str = 'Row %d: ' % (row_index + 1)
            row = self[row_index]

            for col_index in range(len(row)):
                row_str = row_str + '%6.3f  ' % (row[col_index])

            return_str = return_str + '\n' + row_str

        return(return_str)

    def element(self, row_index, col_index):
        """Returns an element of the matrix indexed by row and column.

        Indices begin with 0.

        Parameters
        ----------
        row_index: int
            The row index.
        col_index: int
            The column index.

        Returns
        -------
        float
            The matrix value.
        """
        return ((self[row_index])[col_index])

    def row(self, row_index):
        """Returns a specified row of the matrix.

        Parameters
        ----------
        row_index: int
            The row index.

        Returns
        -------
        list
            The row values.
        """

        return(self[row_index])

    def column(self, col_index):
        """Returns a specified column of the matrix as a numeric list.

        Parameters
        ----------
        col_index: int
            The column index.

        Returns
        -------
        list
            The column values.
        """
        return(NumericList([row[col_index] for row in self]))

    def num_rows(self):
        """Returns the number of rows in the matrix."""

        return(len(self))

    def num_cols(self):
        """Returns the number of columns in the matrix."""

        return (len(self[0]))  # assumes all rows of equal length

    def get_cols(self):
        """Returns list of all columns in a matrix."""
        rng = range(0, self.num_cols())
        return ([self.column(col_index) for col_index in rng])

    def __mul__(m1, m2):
        """Multiplies two Matrix objects and returns the resulting matrix.

        Number of rows in m1 must equal the number of columns in m2.

        Parameters
        ----------
        m1: Matrix
            The first matrix.
        m2: Matrix
            The second matrix.

        Returns
        -------
        Matrix
            The resultant matrix.
        """
        result_rows = []

        # Iterate over the rows in m1.  The first column of row i is formed by
        # multiplying the ith row of m1 by the first column of m2.  The second
        # column is formed by muliplying the ith row of m1 by the second
        # column of m2, etc.
        for row in m1:
            new_row = []

            for col in m2.get_cols():
                new_row.append(row * col)

            result_rows.append(new_row)

        return (Matrix(result_rows))


class Vector:
    "Class to encapsulate vector data and operations."

    def __init__(self, x=0.0, y=0.0, z=0.0):
        """Constructor for a three-dimensional vector.

        Note that two-dimensional vectors can be constructed by omitting one
        of the coordinates, which will default to 0.

        Parameters
        ----------
        x: float
            The x coordinate.
        y: float
            The y coordinate.
        z: float
            The z coordinate.
        """
        self.x = x     # Cartesian x coordinate
        self.y = y     # Cartesian y coordinate
        self.z = z     # Cartesian z coordinate

    def __str__(self):
        """Returns a string representation of the vector."""
        return('Vector: x: %.3f, y: %.3f, z: %.3f' % (self.x, self.y, self.z))

    def set_eq(self, x=None, y=None, z=None):
        """Assigns new value to vector.

        Arguments are now optional to permit this to be used with 2D vectors
        or to modify any subset of coordinates.

        Parameters
        ----------
        x: float
            The x coordinate.
        y: float
            The y coordinate.
        z: float
            The z coordinate.
        """
        if x is not None:
            self.x = x
        if y is not None:
            self.y = y
        if z is not None:
            self.z = z

    def length(self):
        """Returns magnitude of the vector """
        return(sqrt(self.x * self.x + self.y * self.y + self.z * self.z))

    def normalize(self):
        """Returns copy of the normalized vector """
        mag = self.length()
        return (Vector(self.x / mag, self.y / mag, self.z / mag))

    def __mul__(self, rs):
        """Implements Vector * scalar.  Can then use '*' syntax in multiplying
        a vector by a scalar rs

        Parameters
        ----------
        rs: float
            The scalar to multiply.

        Returns
        -------
        Vector
            The resultant vector.
        """
        x = self.x * rs
        y = self.y * rs
        z = self.z * rs
        return (Vector(x, y, z))

    def __rmul__(self, ls):
        """Implements float * Vector.

        Parameters
        ----------
        ls: float
            The scalar to multiply.

        Returns
        -------
        Vector
            The resultant vector.
        """
        x = self.x * ls
        y = self.y * ls
        z = self.z * ls
        return (Vector(x, y, z))

    def __add__(self, rs):
        """Implements Vector + Vector

        Parameters
        ----------
        rs: float
            The scalar to add.

        Returns
        -------
        Vector
            The resultant vector.
        """
        x = self.x + rs.x
        y = self.y + rs.y
        z = self.z + rs.z
        return (Vector(x, y, z))

    def __sub__(self, rs):
        """Implements Vector - Vector.

        Parameters
        ----------
        rs: float
            The scalar to subtract.

        Returns
        -------
        Vector
            The resultant vector.
        """
        x = self.x - rs.x
        y = self.y - rs.y
        z = self.z - rs.z
        return (Vector(x, y, z))

    def __truediv__(self, rs):
        """Implements Vector / float.

        Parameters
        ----------
        rs: float
            The scalar to divide.

        Returns
        -------
        Vector
            The resultant vector.
        """
        x = self.x / rs
        y = self.y / rs
        z = self.z / rs
        return (Vector(x, y, z))

    def __imul__(self, rs):
        """Implements Vector *= float

        Parameters
        ----------
        rs: float
            The scalar to multiply.

        Returns
        -------
        Vector
            The resultant vector.
        """
        self.x *= rs
        self.y *= rs
        self.z *= rs
        return (self)

    def __iadd__(self, rs):
        """Implements Vector += vector.

        Parameters
        ----------
        rs: float
            The scalar to add.

        Returns
        -------
        Vector
            The resultant vector.
        """
        self.x += rs.x
        self.y += rs.y
        self.z += rs.z
        return (self)

    def __isub__(self, rs):
        """Implements Vector -= vector.

        Parameters
        ----------
        rs: float
            The scalar to subtract.

        Returns
        -------
        Vector
            The resultant vector.
        """
        self.x -= rs.x
        self.y -= rs.y
        self.z -= rs.z
        return (self)

    def __idiv__(self, rs):
        """Implements Vector /= float.

        Parameters
        ----------
        rs: float
            The scalar to divide.

        Returns
        -------
        Vector
            The resultant vector.
        """
        self.x /= rs
        self.y /= rs
        self.z /= rs
        return (self)

    def create_matrix(self):
        """Converts a Vector into a single-column matrix."""

        column = [self.x, self.y, self.z]
        return(Matrix([[element] for element in column]))  # singleton list

    # Recommend deletion -- better to use a single interface that takes
    # two vectors.
    def dot(self, V2):
        """returns dot product between two vectors.

        Parameters
        ----------
        V2: Vector
            The vector to dot.

        Returns
        -------
        Vector
            The resultant vector.
        """
        return self.x * V2.x + self.y * V2.y + self.z * V2.z

    # Recommend deletion in favor of non-method version.
    def cross(self, V1, V2):
        """returns cross product of two vectors

        Parameters
        ----------
        V1: Vector
            The vector to cross.
        V2: Vector
            The vector to cross.

        Returns
        -------
        Vector
            The resultant vector.
        """
        x = self.y * V2.z - V1.z * V2.y
        y = self.z * V2.x - V1.x * V2.z
        z = self.x * V2.y - V1.y * V2.x
        return Vector(x, y, z)

    # Replace by separation - RLH
    def angle(self, V2):
        """Returns angle between the two vectors in degrees.

        Parameters
        ----------
        V2: Vector
            The vector to measure.

        Returns
        -------
        float
            The angle between the two vectors.
        """
        R1 = self.length()
        R2 = V2.length()
        adot = dot(self, V2)
        adot = adot / R1 / R2
        adot = min(1., adot)
        adot = max(-1., adot)
        return math2.acosd(adot)

    # RLH: What do these add?  We're creating methods just to access
    # individual attributes.
    def rx(self):
        """The magnitude of x"""
        return self.x

    def ry(self):
        """The magnitude of y"""
        return self.y

    def rz(self):
        """The magnitude of z"""
        return self.z

    # RLH: Suggest deletion in favor of __str__, which has the advantage
    # that it is called on print.
    def display(self):
        """Print the values"""
        return "[%f, %f, %f]" % (self.x, self.y, self.z)

    # RLH: Not necessary if CelestialVector is used.
    def set_xyz(self, ra, dec):
        """Creates a unit vector from spherical coordinates

        Parameters
        ----------
        ra: float
            The right ascension.
        dec: float
            The declination.
        """
        self.x = math2.cosd(dec) * math2.cosd(ra)
        self.y = math2.cosd(dec) * math2.sind(ra)
        self.z = math2.sind(dec)


class CelestialVector (Vector):
    "Class to encapsulate a unit vector on the celestial sphere."

    def __init__(self, ra=0.0, dec=0.0, frame='eq', degrees=True):
        """Constructor for a celestial vector.

        There are two spherical coordinates, a longitudinal coordinate (called
        right ascension), and a latitudinal coordinate (called declination).
        The RA is defined as the counterclockwise angle from a reference
        direction on the equatorial plane; it ranges from 0-360 degrees.
        The DEC is the angle between the vector and the equatorial plane;
        it ranges from -90 to 90 degrees. Angles are specified in degrees but
        represented internally as radians.

        The frame attribute indicates the coordinate frame of the vector,
        which may be 'eq' (equatorial, default), 'ec' (ecliptic), or 'gal'
        (galactic).  In equatorial coordinates, the equatorial plane is the
        celestial equator (extension of the Earth's equator) and the reference
        axis is the vernal equinox.  In ecliptic coordiantes, the equatorial
        plane is the ecliptic (the Earth's orbital plane) and the reference
        axis is usually defined relative to the Sun.  In galactic coordinates,
        the equatorial plane is the plane of the Galaxy.

        The degrees attribute should be True if the RA, DEC inputs are in
        degrees. Otherwise radians is assumed.

        The coordinates "ra" and "dec" may be used in all three systems.
        Other names for coordinates in different frames may be defined for
        clarity.

        A CelestialVector is also an ordinary unit vector, with Cartesian
        coordinates defined relative to the equatorial plane.

        Parameters
        ----------
        ra: float
            The right ascension.
        dec: float
            The declination.
        frame: str
            The frame to use.
        degrees: bool
            Use degrees.
        """
        if (degrees):
            ra = math2.D2R * ra
            dec = math2.D2R * dec

        self.ra = ra
        self.dec = dec
        self.frame = frame

        # Initialize standard vector with translated Cartesian coordinates
        x = cos(ra) * cos(dec)
        y = sin(ra) * cos(dec)
        z = sin(dec)
        Vector.__init__(self, x=x, y=y, z=z)

    def __str__(self, verbose=True):
        """Returns a string representation of the vector.  Displays angles
        in degrees.

        Parameters
        ----------
        verbose: bool
            Print some information.
        """
        a = (math2.R2D * self.ra, math2.R2D * self.dec, self.frame)
        celest_info = 'CelestialVector: RA: %.3fD, DEC: %.3fD, frame: %s' % a

        if (verbose):
            clss = super(CelestialVector, self).__str__()
            celest_info = celest_info + '\n' + clss
            return celest_info

    def set_eq(self, ra, dec, degrees=False):
        """Modifies a celestial vector with a new RA and DEC.

        degrees = True if units are degrees.  Default is radians.

        Parameters
        ----------
        ra: float
            The right ascension.
        dec: float
            The declination.
        degrees: bool
            Use degrees.
        """
        if (degrees):
            ra = math2.D2R * ra
            dec = math2.D2R * dec

            self.ra = ra
            self.dec = dec

        # Update Cartesian coordinates as well.
        x = cos(ra) * cos(dec)
        y = sin(ra) * cos(dec)
        z = sin(dec)
        super(CelestialVector, self).set_eq(x, y, z)

    def update_cartesian(self, x=None, y=None, z=None):
        """Modifies a celestial vector by specifying new Cartesian coordinates.

        Any subset of the Cartesian coordinates may be specifed.

        Parameters
        ----------
        x: float
            The extent in x.
        y: float
            The extent in y.
        z: float
            The extent in z.
        """

        if x is not None:
            self.x = x
        if y is not None:
            self.y = y
        if z is not None:
            self.z = z

        self.ra = atan2(self.y, self.x)  # RA is arctan of y/x
        if (self.ra < 0):                 # Make sure RA is positive
            self.ra += 2 * pi

            self.dec = math2.asin2(self.z)  # DEC is arcsin of z

    def rotate_about_axis(self, angle, axis):
        """This rotates a vector about an axis by the specified angle
        by using a rotation matrix.
        A new vector is returned.

        Axis must be 'x', 'y', or 'z'.
        The x-rotation rotates the y-axis toward the z-axis.
        The y-rotation rotates the z-axis toward the x-axis.
        The z-rotation rotates the x-axis toward the y-axis.

        Parameters
        ----------
        angle: float
            The angle of rotation.
        axis: str
            The axis to rotate about, ['x', 'y', 'z'].

        Returns
        -------
        result : Vector
            The rotated vector.
        """
        if (axis == 'x'):
            rot_matrix = Matrix([[1, 0, 0], [0, cos(angle), -sin(angle)],
                                 [0, sin(angle), cos(angle)]])

        elif (axis == 'y'):
            rot_matrix = Matrix([[cos(angle), 0, sin(angle)], [0, 1, 0],
                                 [-sin(angle), 0, cos(angle)]])

        elif (axis == 'z'):
            rot_matrix = Matrix([[cos(angle), -sin(angle), 0],
                                 [sin(angle), cos(angle), 0], [0, 0, 1]])

        else:
            print('Error')
            return

        new_matrix = rot_matrix * self.create_matrix()
        new_vector = new_matrix.column(0)
        result = CelestialVector()  # initialize with Cartesian coordiantes
        result.update_cartesian(x=new_vector[0], y=new_vector[1],
                                z=new_vector[2])
        return result

    def rotate_about_eigenaxis(self, angle, eigenaxis):
        """Rotates a vector about arbitrary eigenaxis.

        eigenaxis = Vector object (axis about which to rotate).
        angle = angle to rotate by in radians.
        Rotation is counterclockwise looking outward from origin along
        eigenaxis. Function uses rotation matrix from Rodrigues formula.

        Note: This function is more general than rotate_about_axis above and
        could be used in its place.  However, rotate_about_axis is faster and
        clearer when the rotation axis is one of the Cartesian axes.

        Parameters
        ----------
        angle: float
            The angle of rotation.
        eigenaxis: Vector
            The eigenaxis to rotate about.

        Returns
        -------
        result : Vector
            The rotated vector.
        """
        cos_ang = cos(angle)    # Used repeatedly below
        sin_ang = sin(angle)

        # Fill out the Rodrigues rotation matrix
        R11 = cos_ang + eigenaxis.x**2 * (1 - cos_ang)
        R12 = eigenaxis.x * eigenaxis.y * (1 - cos_ang) - eigenaxis.z * sin_ang
        R13 = eigenaxis.x * eigenaxis.z * (1 - cos_ang) + eigenaxis.y * sin_ang
        R21 = eigenaxis.x * eigenaxis.y * (1 - cos_ang) + eigenaxis.z * sin_ang
        R22 = cos_ang + eigenaxis.y**2 * (1 - cos_ang)
        R23 = eigenaxis.y * eigenaxis.z * (1 - cos_ang) - eigenaxis.x * sin_ang
        R31 = eigenaxis.x * eigenaxis.z * (1 - cos_ang) - eigenaxis.y * sin_ang
        R32 = eigenaxis.y * eigenaxis.z * (1 - cos_ang) + eigenaxis.x * sin_ang
        R33 = cos_ang + eigenaxis.z**2 * (1 - cos_ang)

        r1, r2, r3 = [R11, R12, R13], [R21, R22, R23], [R31, R32, R33]
        rot_matrix = Matrix([r1, r2, r3])
        new_matrix = rot_matrix * self.create_matrix()
        new_vector = new_matrix.column(0)
        result = CelestialVector()  # initialize with Cartesian coordinates
        result.update_cartesian(x=new_vector[0], y=new_vector[1],
                                z=new_vector[2])
        return(result)

    def rotate_using_quaternion(self, angle, eigenaxis):
        """Rotates a vector about arbitrary eigenaxis using quaternion.

        This is an alternative formulation for rotate_about_eigenaxis.
        Interface is the same as rotate_about_eigenaxis.

        Parameters
        ----------
        angle: float
            The angle of rotation.
        eigenaxis: Vector
            The eigenaxis to rotate about.

        Returns
        -------
        Vector
            The rotated vector.
        """
        q = Quaternion(eigenaxis, 0.0)

        # Need to negate here because set_values performs a negative rotation
        # quaternion now represents the rotation
        q.set_values(eigenaxis, -angle)
        return(make_celestial_vector(q.cnvrt(self)))

    def transform_frame(self, new_frame):
        """Transforms coordinates between celestial and ecliptic frames
        and returns result as a new CelestialVector.
        If new coordinate frame is the same as the old, a copy of the vector
        is returned.

        Parameters
        ----------
        new_frame: str
            Convert to new frame.

        Returns
        -------
        result : Vector
            The transformed vector.
        """
        result = None
        gal_ec = new_frame == 'gal' and self.frame == 'ec'
        ec_gal = new_frame == 'ec' and self.frame == 'gal'

        # Equatorial to ecliptic: rotate z-axis toward y-axis.
        if ((new_frame == 'ec') and (self.frame == 'eq')):
            result = self.rotate_about_axis(-math2.OBLIQUITY, 'x')

        # Ecliptic to equatorial: rotate y-axis toward z-axis.
        elif ((new_frame == 'eq') and (self.frame == 'ec')):
            result = self.rotate_about_axis(math2.OBLIQUITY, 'x')

        elif ((new_frame == 'gal') and (self.frame == 'eq')):
            # Use formula from Wayne Kinzel's book, adjusted for
            # J2000 coordinates.
            a = cos(self.dec) * cos(NGP.longitude) * \
                cos(self.ra - NGP.latitude)
            b = math2.asin2(a + sin(self.dec) * sin(NGP.longitude))
            arg1 = sin(self.dec) - sin(b) * sin(NGP.longitude)
            arg2 = cos(self.dec) * sin(self.ra - NGP.latitude) * \
                cos(NGP.longitude)

            lng = atan2(arg1, arg2) + NGP.anode

            result = CelestialVector(lng, b, degrees=False)

        elif ((new_frame == 'eq') and (self.frame == 'gal')):
            lng = self.ra   # use l, b notation here for clarity
            b = self.dec
            term1 = cos(b) * cos(NGP.longitude) * sin(lng - NGP.anode)
            dec = math2.asin2(term1 + sin(b) * sin(NGP.longitude))
            arg1 = cos(b) * cos(lng - NGP.anode)
            sinterm = sin(NGP.longitude) * sin(lng - NGP.anode)
            arg2 = sin(b) * cos(NGP.longitude) - cos(b) * sinterm
            ra = atan2(arg1, arg2) + NGP.latitude

            result = CelestialVector(ra, dec, degrees=False)

        elif gal_ec or ec_gal:
            print("""Error: Direct conversion between ecliptic and\
                     galactic coordinates not supported yet""")

        elif new_frame != self.frame:
            print("Error: unrecognized coordinate frame.")

        # If there was an error, return a copy of the initial vector.
        if result is None:
            result = CelestialVector(self.ra, self.dec, self.frame, False)

        else:
            result.frame = new_frame  # record new frame

        return (result)

    def rotate_by_posang(self, pa):
        """Returns the vector that results from rotating the self vector
        counterclockwise from the North projection onto the plane
        orthogonal to that vector by the specified position angle
        (in radians). See "V3-axis Position Angle", John Isaacs, May 2003 for
        further discussion.

        Parameters
        ----------
        pa: float
            The position angle.

        Returns
        -------
        result : Vector
            The rotated vector.
        """
        x_coord = -cos(self.ra) * sin(self.dec) * \
            cos(pa) - sin(self.ra) * sin(pa)
        y_coord = -sin(self.ra) * sin(self.dec) * \
            cos(pa) + cos(self.ra) * sin(pa)
        z_coord = cos(self.dec) * cos(pa)
        result = CelestialVector()
        result.update_cartesian(x_coord, y_coord, z_coord)
        return(result)

    def position_angle(self, v):
        """Returns the position angle of v at the self vector, in radians.

        v is an arbitrary vector that should be a CelestialVector object.
        The position angle is the angle between the North vector on the
        plane orthogonal to the self vector and the projection of v onto
        that plane, defined counterclockwise.
        See "V3-axis Position Angle", John Isaacs, May 2003 for
        further discussion.

        Parameters
        ----------
        v: Vector
            The vector to measure against.

        Returns
        -------
        pa : float
            The position angle between the two vectors.
        """
        y_coord = cos(v.dec) * sin(v.ra - self.ra)
        b = cos(v.dec) * sin(self.dec) * cos(v.ra - self.ra)
        x_coord = sin(v.dec) * cos(self.dec) - b
        pa = atan2(y_coord, x_coord)

        if (pa < 0):
            pa += (2 * pi)  # PA has range 0-360 degrees

        return(pa)


class Attitude(CelestialVector):
    "Defines an Observatory attitude by adding a position angle."""

    def __init__(self, ra=0.0, dec=0.0, pa=0.0, frame='eq', degrees=True):
        """Constructor for an Attitude.

        pa = position_angle in degrees(default) or radians if degrees=False
        is specified. Other arguments are the same as with CelestialVector

        Parameters
        ----------
        ra: float
            The right ascension.
        dec: float
            The declination.
        pa: float
            The position angle.
        frame: str
            The frame to use.
        degrees: bool
            Use degrees.
        """
        super(Attitude, self).__init__(ra=ra, dec=dec, frame=frame,
                                       degrees=degrees)

        if (degrees):   # convert into radians
            pa = math2.D2R * pa

        self.pa = pa

    def __str__(self, verbose=True):
        """Returns a string representation of the attitude.

        verbose (optional) = flag indicating whether detailed Vector
        information should be included.

        Parameters
        ----------
        verbose: bool
            Print information.

        Returns
        -------
        att_info : str
            A string representation of the attitute.
        """
        att_info = 'Attitude: PA: %.3fD' % (math2.R2D * self.pa)
        att_info = att_info + '\n' + super(Attitude, self).__str__(verbose)
        return att_info


# Functions that operate on vectors but are not methods.
def dot(v1, v2):
    """returns dot product between two vectors, non class member.

    Parameters
    ----------
    v1: Vector
        The first vector.
    v2: Vector
        The second vector.

    Returns
    -------
    float
        The dot product of the vectors.
    """
    return(v1.x * v2.x + v1.y * v2.y + v1.z * v2.z)


def cross(v1, v2):
    """Returns cross product between two vectors, non class member.

    Parameters
    ----------
    v1: Vector
        The first vector.
    v2: Vector
        The second vector.

    Returns
    -------
    float
        The cross product of the vectors.
    """
    x = v1.y * v2.z - v1.z * v2.y
    y = v1.z * v2.x - v1.x * v2.z
    z = v1.x * v2.y - v1.y * v2.x
    return Vector(x, y, z)


def separation(v1, v2, norm=False):
    """Returns angle between two unit vectors in radians.

    The angle between two normalized vectors is the arc-cosine of the dot
    product. Unless the norm attribute is set to True, it is assumed the
    vectors are already normalized (for performance).

    Parameters
    ----------
    v1: Vector
        The first vector.
    v2: Vector
        The second vector.
    norm: bool
        Normalize the vectors.

    Returns
    -------
    separation : float
        The separation of the vectors.
    """
    if (norm):
        v1 = v1.normalize()
        v2 = v2.normalize()

    separation = math2.acos2(dot(v1, v2))

    # For very small angles, cos and acos behave poorly as the cosine of a
    # very small angle is interpreted as 1.0.  Therefore, recompute using
    # the cross product if the result is less than 1 degree.
    if (separation < math2.D2R):
        vcross = cross(v1, v2)
        separation = math2.asin2(vcross.length())

    return(separation)


def ra_delta(v1, v2):
    """Returns difference in right ascension between two CelestialVectors.

    Parameters
    ----------
    v1: Vector
        The first vector.
    v2: Vector
        The second vector.

    Returns
    -------
    delta_ra : float
        The difference in RA.
    """
    delta_ra = v1.ra - v2.ra

    # Check for zero crossings.  If the difference exceeds 180 degrees,
    # adjust by 360 in opposite direction.

    if (delta_ra < -pi):
        delta_ra = delta_ra + 2 * pi
    elif (delta_ra > pi):
        delta_ra = delta_ra - 2 * pi

    return(delta_ra)


def ra_separation(v1, v2):
    """Returns separation in right ascension between two CelestialVectors.
    This is accurate only if the difference in declination is small.

    |sep| = DELTA-RA cos DEC

    Parameters
    ----------
    v1: Vector
        The first vector.
    v2: Vector
        The second vector.

    Returns
    -------
    float
        The separation between RA values.
    """
    delta_ra = ra_delta(v1, v2)
    dec = math2.avg2(v1.dec, v2.dec)  # use average of the two declinations.
    return(delta_ra * cos(dec))


def dec_separation(v1, v2):
    """Returns difference in declination between two CelestialVectors.

    Parameters
    ----------
    v1: Vector
        The first vector.
    v2: Vector
        The second vector.

    Returns
    -------
    float
        The separation between Dec values
    """
    return(v1.dec - v2.dec)    # simply take the difference in declination


def make_celestial_vector(v):
    """Takes a Vector object and creates an equivalent CelestialVector.

    Input vector v must be a unit vector.

    Parameters
    ----------
    v: Vector
        The vector to convert.

    Returns
    -------
    result : Vector
        The updated vector.
    """
    result = CelestialVector()
    result.update_cartesian(v.x, v.y, v.z)
    return(result)


def projection(v, axis):
    """Returns projection of vector v on plane normal to axis.

    First take cross-product of v and the axis and normalize it.
    Then cross the axis with the result and return a CelestialVector.
    See http://www.euclideanspace.com/maths/geometry/elements/plane/
    lineOnPlane/index.htm.

    Parameters
    ----------
    v: Vector
        The vector to convert.
    axis: str
        The axis to project onto.

    Returns
    -------
    Vector
        The updated vector.
    """
    return(make_celestial_vector(cross(axis, (cross(v, axis)).normalize())))


def pos_V_to_ra_dec(V):
    """Returns tuple of spherical angles from unit direction Vector.

    Parameters
    ----------
    V: Vector
        The vector to analyze.

    Returns
    -------
    ra : float
        The ra of the vector.
    dec : float
        The dec of the vector.
    """
    ra = math2.atan2d(V.y, V.x)
    V.z = min(1., V.z)
    V.z = max(-1., V.z)
    dec = math2.asind(V.z)
    if ra < 0.:
        ra += 360.
    return(ra, dec)


# RLH: Recommend replacement by separation.
def angle(V1, V2):
    """returns angle between two vectors in degrees, non class member.

    Parameters
    ----------
    V1: Vector
        The first vector.
    V2: Vector
        The second vector.

    Returns
    -------
    float
        The angle between the vectors.
    """
    R1 = V1.length()
    R2 = V2.length()
    adot = dot(V1, V2)
    adot = adot / R1 / R2
    adot = min(1., adot)
    adot = max(-1., adot)
    return math2.acosd(adot)


def vel_ab(U, Vel):
    """Takes a unit vector and a velocity vector(km/s) and returns a unit
    vector modidifed by the velocity abberation.

    Parameters
    ----------
    U: Vector
        The unit vector.
    Vel: Vector
        The velocity vector to multiply.

    Returns
    -------
    Vector
        The modified vector.
    """
    c = 2.9979e5  # speed of light in km/s
    Beta = Vel * (1. / c)
    rgamma = sqrt(1. - dot(Beta, Beta))  # This is 1/gamma
    ubeta = dot(U, Beta)
    b = (1. / (1. + ubeta))
    return (U * rgamma + Beta * (1. + (1. - rgamma) * ubeta / dot(Beta, Beta))) * b



# ! /usr/bin/env python
# Module ephemeris.py
import math
import sys
import time

from . import astro_funcx as astro_func
from . import quaternionx as qx
from . import time_extensionsx as time2

D2R = math.pi / 180.  # degrees to radians
R2D = 180. / math.pi  # radians to degrees
PI2 = 2. * math.pi  # 2 math.pi
MIN_SUN_ANGLE = 84.8 * D2R  # minimum Sun angle, in radians
MAX_SUN_ANGLE = 135.0 * D2R  # maximum Sun angle, in radians
SUN_ANGLE_PAD = 0.5 * D2R   # pad away from Sun angle limits

Qecl2eci = qx.QX(23.439291 * D2R)  # At J2000 equinox


class Ephemeris:
    """A class for the ephemeris of an observation.

    History
    -------
    07/31/2008 Added functions for sun position
    12/05/2008 Switched to spline fit.
    07/29/2010 Added OP window calc
    08/03/2010 Got rid of degrees trig functions
              Removed in_FOR, is_valid_att etc. as those are S/C dependent
    """

    def __init__(self, ephem_file, cnvrt=False):
        """Ephemeris constructor

        Parameters
        ----------
        ephem_file: str
            The path to the ephemeris file.
        cnvrt: bool, optional
            Converts into Ecliptic frame.
        """
        if cnvrt:
            print("Using Ecliptic Coordinates")
        else:
            print("Using Equatorial Coordinates")
        self.datelist = []
        self.xlist = []
        self.ylist = []
        self.zlist = []
        self.amin = 0.
        self.amax = 0.
        aV = qx.Vector(0., 0., 0.)
        if ephem_file.find("l2_halo_FDF_060619.trh") > -1:
            ascale = 0.001
        else:
            ascale = 1.0
        fin = open(ephem_file, 'r')
        finLines = fin.readlines()
        for item in finLines[2:]:
            item = item.strip()
            item = item.split()
            adate = time2.mjd_from_string(item[0])  # represent dates as mjds
            x = float(item[1]) * ascale
            y = float(item[2]) * ascale
            z = float(item[3]) * ascale
            if cnvrt:
                aV.set_eq(x, y, z)
                ll = aV.length()
                aV = aV / ll
                aV = Qecl2eci.inv_cnvrt(aV)
                aV = aV * ll
                x = aV.rx()
                y = aV.ry()
                z = aV.rz()
            self.datelist.append(adate)
            self.xlist.append(x)
            self.ylist.append(y)
            self.zlist.append(z)

            if self.amin == 0.:
                self.amin = adate
        self.amax = adate
        fin.close()

    def report_ephemeris(self, limit=100000, pathname=None):
        """Prints a formatted report of the ephemeris.

        Parameters
        ----------
        limit: int, optional
            The number of records to report.
        pathname: str, optional
            The path to a file to hold the report.
        """
        num_to_report = min(limit, len(self.datelist))

        if (pathname):
            dest = open(pathname, 'w')
            print(('# Generated %s\n' % (time.ctime())), file=dest)
        else:
            dest = sys.stdout  # defaults to standard output

        print(('%17s  %14s  %14s  %14s\n' %
               ('DATE      ', 'X (KM)   ', 'Y (KM)   ', 'Z (KM)   ')), file=dest)

        for num in range(num_to_report):
            date = self.datelist[num]
            x = self.xlist[num]
            y = self.ylist[num]
            z = self.zlist[num]

            fmt = time2.display_date(date), x, y, z
            print(('%17s  %14.3f  %14.3f  %14.3f' % fmt), file=dest)

        if (pathname):
            dest.close()   # Clean up

    def pos(self, adate):
        """Computes the position of the telescope at a given date using the
        grid of positions of the ephemeris as a starting point and
        applying a linear interpolation between the ephemeris grid points

        Parameters
        ----------
        adate: datetime.datetime object
            The date of the observation.

        Returns
        -------
        qx.Vector object
            The position of the telescope as a Vector object
        """
        cal_days = adate - self.datelist[0]
        index = int(cal_days)
        if (index == len(self.datelist) - 1):
            index = index - 1
        frac = cal_days - index
        x = (self.xlist[index + 1] - self.xlist[index]) * \
            frac + self.xlist[index]
        y = (self.ylist[index + 1] - self.ylist[index]) * \
            frac + self.ylist[index]
        z = (self.zlist[index + 1] - self.zlist[index]) * \
            frac + self.zlist[index]
        return qx.Vector(x, y, z)

    def Vsun_pos(self, adate):
        """The vector of the sun at the given date

        Parameters
        ----------
        adate: datetime
            The date of the observation

        Returns
        -------
        Vector
            The position of the sun as a Vector object
        """
        Vsun = -1. * self.pos(adate)
        Vsun = Vsun / Vsun.length()
        return Vsun

    def sun_pos(self, adate):
        """The coordinates of the sun at the given date

        Parameters
        ----------
        adate: datetime.datetime object
            The date of the observation.

        Returns
        -------
        tuple
            The coordinates of the sun.
        """
        Vsun = -1. * self.pos(adate)
        Vsun = Vsun / Vsun.length()
        coord2 = math.asin(unit_limit(Vsun.z))
        coord1 = math.atan2(Vsun.y, Vsun.x)
        if coord1 < 0.:
            coord1 += PI2
        return (coord1, coord2)

    def normal_pa(self, adate, tgt_c1, tgt_c2):
        """Calculate the V3 position

        Parameters
        ----------
        adate: datetime.datetime object
            The date of the observation.
        tgt_c1: float
            The RA in radians.
        tgt_c2: float
            The Dec in radians.

        Returns
        -------
        float
            The V3 position.
        """
        (sun_c1, sun_c2) = self.sun_pos(adate)
        sun_pa = astro_func.pa(tgt_c1, tgt_c2, sun_c1, sun_c2)
        V3_pa = sun_pa + math.pi  # We want -V3 pointed towards sun.
        if V3_pa < 0.:
            V3_pa += PI2
        if V3_pa >= PI2:
            V3_pa -= PI2
        return V3_pa

    def long_term_attitude(self, date):
        """Defines a long-term safe attitude as of a given date.

        Parameters
        ----------
        date: float
             The date of computation, as an mjd.

        Returns
        -------
        Attitude
            The Attitude object at the given date.
        """
        # Retrieve Sun's position and transform to ecliptic coordinates.
        (sun_ra, sun_dec) = self.sun_pos(date)   # RA range 0-PI2
        vSun = qx.CelestialVector(sun_ra, sun_dec, degrees=False)
        vSun = vSun.transform_frame('ec')

        # Now subtract the minimum Sun angle plus a pad from the
        # ecliptic longitude.
        # Sun angle steadily decreases as the Earth (and JWST with it)
        # revolve counterclockwise, so this should maximize the duration
        # when the attitude is within the FOR.  Set the latitude to
        # 0 (ecliptic).
        # Normalize to 0-360 degrees, and convert back into equatorial
        # coordinates.
        # Then set the normal PA, which should be valid at the ecliptic for
        # the duration of the visibility window.
        longitude = vSun.ra - MIN_SUN_ANGLE - SUN_ANGLE_PAD

        if (longitude < 0):
            longitude = longitude + PI2

        vec1 = qx.CelestialVector(longitude, 0.0, frame='ec', degrees=False)
        vec1 = vec1.transform_frame('eq')
        pa = self.normal_pa(date, vec1.ra, vec1.dec)
        return(qx.Attitude(vec1.ra, vec1.dec, pa, degrees=False))

    def is_valid(self, date, ngc_1, ngc_2, V3pa):
        """Indicates whether an attitude is valid at a given date.

        Parameters
        ----------
        date: float
            The date of the observation.
        ngc_1: flaot
            The RA of the reference in radians.
        ngc_2: float
            The Dec of the reference in radians.
        V3pa: float
            The V3 position of the telescope.

        Returns
        -------
        bool
            Is it a valid PA.
        """
        # First check that the date is within the time interval of
        # the ephemeris.
        if ((date < self.amin) or (date > self.amax)):
            return False

        (sun_1, sun_2) = self.sun_pos(date)
        d = astro_func.dist(ngc_1, ngc_2, sun_1, sun_2)
        vehicle_pitch = math.pi / 2 - d   # see JI memo from May 2006
        # sun pitch is always equal or greater than sun angle (V1 to sun)
        if (d < MIN_SUN_ANGLE or d > MAX_SUN_ANGLE):
            return False
        # now checking the roll and pitch angle combination
        pa = astro_func.pa(ngc_1, ngc_2, sun_1, sun_2) + math.pi
        roll = math.acos(math.cos(V3pa - pa))
        sun_roll = math.asin(math.sin(roll) * math.cos(vehicle_pitch))
        if (abs(sun_roll) <= 5.2 * D2R):
            sun_pitch = math.atan2(math.tan(vehicle_pitch), math.cos(roll))
            if (sun_pitch <= 5.2 * D2R and sun_pitch >= -45. * D2R):
                return True
        return False

    def in_FOR(self, date, ngc_1, ngc_2):
        """Test if in the FOR

        Parameters
        ----------
        date: float
            The date of the observation.
        ngc_1: flaot
            The RA of the reference in radians.
        ngc_2: float
            The Dec of the reference in radians.

        Returns
        -------
        bool
            Is it in the FOR.
        """
        (sun_1, sun_2) = self.sun_pos(date)
        d = astro_func.dist(ngc_1, ngc_2, sun_1, sun_2)
        # sun pitch is always equal or greater than sun angle (V1 to sun)
        if (d < MIN_SUN_ANGLE or d > MAX_SUN_ANGLE):
            return False
        return True

    def bisect_by_FOR(self, in_date, out_date, ngc_1, ngc_2):
        """Find the midpoint in time between in and out of FOR,
        assumes only one "root" in interval

        Parameters
        ----------
        in_date: float
            The in date of the observation.
        out_date: float
            The out date of the observation.
        ngc_1: flaot
            The RA of the reference in radians.
        ngc_2: float
            The Dec of the reference in radians.

        Returns
        -------
        float
            The midpoint in time.
        """
        delta_days = 200.
        mid_date = (in_date + out_date) / 2.
        while delta_days > 0.000001:
            (sun_1, sun_2) = self.sun_pos(mid_date)
            d = astro_func.dist(ngc_1, ngc_2, sun_1, sun_2)
            if (d > MAX_SUN_ANGLE or d < MIN_SUN_ANGLE):
                out_date = mid_date
            else:
                in_date = mid_date
            mid_date = (in_date + out_date) / 2.
            delta_days = abs(in_date - out_date) / 2.
            # print "UU", mid_date
        # ensure returned date always in FOR
        if in_date > out_date:
            mid_date = mid_date + 0.000001
        else:
            mid_date = mid_date - 0.000001
        return mid_date

    def bisect_by_attitude(self, in_date, out_date, ngc_1, ngc_2, pa):
        """Find the midpoint in time between in and out of FOR,
        assumes only one "root" in interval

        Parameters
        ----------
        in_date: float
            The in date of the observation.
        out_date: float
            The out date of the observation.
        ngc_1: flaot
            The RA of the reference in radians.
        ngc_2: float
            The Dec of the reference in radians.
        pa: float
            The position angle.

        Returns
        -------
        float
            The midpoint in time.
        """
        icount = 0
        delta_days = 200.
        mid_date = (in_date + out_date) / 2.
        # print "bisect >", in_date, out_date, abs(in_date-out_date )
        while delta_days > 0.000001:
            if self.is_valid(mid_date, ngc_1, ngc_2, pa):
                in_date = mid_date
            else:
                out_date = mid_date
            mid_date = (in_date + out_date) / 2.
            delta_days = abs(in_date - out_date) / 2.
            # print "UU", mid_date
            icount = icount + 1
        # print " bisected >", icount
        return mid_date

    def OP_window(self, adate, ngc_1, ngc_2, pa, mdelta, pdelta):
        """Attitude at adate must be valid, else returns (0, 0).
        If valid, returns (adate-mdelta, adate+pdelta) or the constraint
        window, which ever is smaller.

        Parameters
        ----------
        date: float
            The date of the observation.
        ngc_1: flaot
            The RA of the reference in radians.
        ngc_2: float
            The Dec of the reference in radians.
        pa: float
            The position angle.
        mdelta: float
            Delta time on low end.
        pdelta: float
            Delta time on high end.

        Returns
        -------
        tuple
            The OP window.
        """
        if self.is_valid(adate, ngc_1, ngc_2, pa):
            if self.is_valid(adate - mdelta, ngc_1, ngc_2, pa):
                OP_min = adate - mdelta
            else:
                OP_min = self.bisect_by_attitude(adate, adate - mdelta,
                                                 ngc_1, ngc_2, pa)
            if self.is_valid(adate + pdelta, ngc_1, ngc_2, pa):
                OP_max = adate + pdelta
            else:
                OP_max = self.bisect_by_attitude(adate, adate + pdelta,
                                                 ngc_1, ngc_2, pa)
        else:
            OP_min = 0.
            OP_max = 0.
        return (OP_min, OP_max)


def unit_limit(x):
    """ Forces value to be in [-1, 1].

    Parameters
    ----------
    x: float, int
        The value to adjust.

    Returns
    -------
    float
        The adjusted value.
    """
    return min(max(-1., x), 1.)



"""UNDER CONSTRUCTION"""

import glob
import os
import pysiaf

import astropy.coordinates as crd
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from astroquery.irsa import Irsa
from matplotlib import cm
from scipy.io import readsav
from astropy.io import fits
from exoctk.utils import get_env_variables

EXOCTK_DATA = os.environ.get('EXOCTK_DATA')
if not EXOCTK_DATA:
    print('WARNING: The $EXOCTK_DATA environment variable is not set. '
          'Contamination overlap will not work. Please set the '
          'value of this variable to point to the location of the exoctk_data '
          'download folder.  Users may retreive this folder by clicking the '
          '"ExoCTK Data Download" button on the ExoCTK website, or by using '
          'the exoctk.utils.download_exoctk_data() function.'
          )
    TRACES_PATH = None
else:
    TRACES_PATH = os.path.join(EXOCTK_DATA,  'exoctk_contam', 'traces')


class fieldSim:
    def __init__(self, RA, DEC, INSTRUMENT):
        self.ra = RA
        self.dec = DEC
        self.instrument = INSTRUMENT

        if 'NIRCam' in self.instrument:
            siaf = pysiaf.Siaf('NIRCam')
            dimX, dimY = 51, 1343
            rad = 2.5
            pixel_scale = siaf.
            xSweet, ySweet = siaf. , siaf.
        elif 'MIRI' in self.instrument:
            siaf = pysiaf.Siaf('MIRI')
            dimX, dimY = 51, 1343
            rad = 2.5
            pixel_scale = siaf.
            xSweet, ySweet = siaf. , siaf.

    def getStars(self, ):



"""Functions to simulate lightcurve data from ExoMAST parameters

Author: Joe Filippazzo
Email: jfilippazzo@stsci.edu
"""
import numpy as np
from bokeh.plotting import figure, show
try:
    import batman
except ImportError:
    print("Could not import batman. Functionality may be limited.")

from .. import utils


def simulate_lightcurve(target, snr=1000., npts=1000, nbins=10, radius=None, ldcs=('quadratic', [0.1, 0.1]), plot=False):
    """Simulate lightcurve data for the given target exoplanet

    Parameters
    ----------
    target: str
        The name of the target to simulate
    snr: float
        The signal to noise to use
    npts: int
        The number of points in each lightcurve
    nbins: int
        The number of lightcurves
    radius: array-like, float (optional)
        The radius or radii value(s) to use
    ldcs: sequence
        The limb darkening profile name and coefficients
    plot: bool
        Plot the figure

    Returns
    -------
    tuple
        The time, flux, uncertainty, and transit parameters
    """
    try:

        # Get the system parameters from ExoMAST
        targ, url = utils.get_target_data(target)
        name = targ.get('canonical_name') or target
        t0 = targ.get('transit_time', 0.)
        dt = targ.get('transit_duration', 1.)

        # Generate transit parameters with batman
        params = batman.TransitParams()
        params.t0 = t0
        params.rp = targ.get('Rp/Rs') or 0.1
        params.per = targ.get('orbital_period') or 0.5
        params.inc = targ.get('inclination') or 90.
        params.a = targ.get('a/Rs') or 15.
        params.ecc = targ.get('eccentricity') or 0.
        params.w = targ.get('omega') or 90.
        params.limb_dark = 'nonlinear' if ldcs[0] == '4-parameter' else ldcs[0]
        params.transittype = 'primary'
        params.u = ldcs[1]

        # Generate a time axis
        time = np.linspace(t0 - dt, t0 + dt, npts)

        # Make the transit model
        transit = batman.TransitModel(params, time, transittype='primary')

        # Generate the lightcurves
        flux = []
        if radius is None:
            radius = params.rp
        radii = [radius] * nbins if isinstance(radius, (int, float)) else radius
        for r in radii:
            params.rp = r
            flux.append(transit.light_curve(params))

        # Add noise
        ideal_flux = np.asarray(flux)
        flux = np.random.normal(loc=ideal_flux, scale=ideal_flux/snr)
        unc = flux - ideal_flux

        # Plot it
        if plot:
            fig = figure(title=name)
            fig.circle(time, flux[0])
            fig.xaxis.axis_label = targ.get('transit_time_unit') or 'MJD'
            fig.yaxis.axis_label = 'Relative Flux'
            show(fig)

        return time, flux, unc, targ

    except Exception:
        raise ValueError('{}: Could not simulate light curve for this target'.format(target))



"""Produces a graph of the visibility & accessible position angles
for a given RA & DEC, and prints out corresponding information,
including the ranges of accessible and inaccessible PAs.

Usage: python visibilityPA.py RA DEC [targetName]
 if targetName is specified, then the figure is saved

-Created by David Lafreniere, March 2016
-makes use of (and hacks) several scripts created by Pierre Ferruit
 that are part of the JWST Python tools JWSTpylib and JWSTpytools
"""
import datetime
import math
import pkg_resources

from astropy.table import Table
from astropy.time import Time
from bokeh.plotting import figure, ColumnDataSource
from bokeh.models import HoverTool, ranges
from bokeh.models.widgets import Panel, Tabs
import matplotlib.dates as mdates
import numpy as np

from . import ephemeris_old2x as EPH
from jwst_gtvt.find_tgt_info import get_table


D2R = math.pi / 180.  # degrees to radians
R2D = 180. / math.pi  # radians to degrees


def checkVisPA(ra, dec, targetName=None, ephFileName=None, fig=None):
    """Check the visibility at a range of position angles.

    Parameters
    ----------
    ra: str
        The RA of the target in hh:mm:ss.s or dd:mm:ss.s or representing a float
    dec: str
        The Dec of the target in hh:mm:ss.s or dd:mm:ss.s or representing a float
    targetName: str
        The target name
    ephFileName: str
        The filename of the ephemeris file
    fig: bokeh.plotting.figure
        The figure to plot to

    Returns
    -------
    paGood : float
        The good position angle.
    paBad : float
        The bad position angle.
    gd : matplotlib.dates object
       The greogrian date.
    fig : bokeh.plotting.figure object
        The plotted figure.

    """
    if ephFileName is None:
        eph_file = 'data/contam_visibility/JWST_ephem_short.txt'
        ephFileName = pkg_resources.resource_filename('exoctk', eph_file)
    if ra.find(':') > -1:  # format is hh:mm:ss.s or  dd:mm:ss.s
        ra = convert_ddmmss_to_float(ra) * 15. * D2R
        dec = convert_ddmmss_to_float(dec) * D2R
    else:  # format is decimal
        ra = float(ra) * D2R
        dec = float(dec) * D2R

    # load ephemeris
    eclFlag = False
    eph = EPH.Ephemeris(ephFileName, eclFlag)

    # convert dates from MJD to Gregorian calendar dates
    mjd = np.array(eph.datelist)
    d = mdates.julian2num(mjd + 2400000.5)
    gd = mdates.num2date(d)

    # loop through dates and determine VIS and PAs (nominal, min, max)
    vis = np.empty(mjd.size, dtype=bool)
    paNom = np.empty(mjd.size)
    paMin = np.empty(mjd.size)
    paMax = np.empty(mjd.size)
    for i in range(mjd.size):

        # is it visible?
        vis[i] = eph.in_FOR(mjd[i], ra, dec)

        # nominal PA at this date
        pa = eph.normal_pa(mjd[i], ra, dec)

        # search for minimum PA allowed by roll
        pa0 = pa
        while eph.is_valid(mjd[i], ra, dec, pa0 - 0.002):
            pa0 -= 0.002

        # search for maximum PA allowed by roll
        pa1 = pa
        while eph.is_valid(mjd[i], ra, dec, pa1 + 0.002):
            pa1 += 0.002

        paNom[i] = (pa * R2D) % 360
        paMin[i] = (pa0 * R2D) % 360
        paMax[i] = (pa1 * R2D) % 360

    # does PA go through 360 deg?
    wrap = np.any(np.abs(np.diff(paNom[np.where(vis)[0]])) > 350)

    # Determine good and bad PA ranges
    # Good PAs
    i, = np.where(vis)
    pa = np.concatenate((paNom[i], paMin[i], paMax[i]))

    if wrap:
        pa = np.append(pa, (0., 360.))
    pa.sort()

    i1, = np.where(np.diff(pa) > 10)
    i0 = np.insert(i1 + 1, 0, 0)
    i1 = np.append(i1, -1)
    paGood = np.dstack((pa[i0], pa[i1])).round(1).reshape(-1, 2).tolist()

    # bad PAs (complement of the good PAs)
    paBad = []
    if paGood[0][0] > 0:
        paBad.append([0., paGood[0][0]])
    for i in range(1, len(paGood)):
        paBad.append([paGood[i - 1][1], paGood[i][0]])
    if paGood[-1][1] < 360.:
        paBad.append([paGood[-1][1], 360.])

    # Make a figure
    if fig is None or fig:
        tools = 'crosshair, reset, hover, save'
        radec = ', '.join([str(ra), str(dec)])
        fig = figure(tools=tools, plot_width=800, plot_height=400,
                     x_axis_type='datetime',
                     title=targetName or radec)

    # Do all figure calculations
    iBad, = np.where(vis == False)
    paMasked = np.copy(paNom)
    paMasked[iBad] = np.nan
    gdMasked = np.copy(gd)

    i = np.argmax(paNom)
    if paNom[i + 1] < 10:
        i += 1
    paMasked = np.insert(paMasked, i, np.nan)
    gdMasked = np.insert(gdMasked, i, gdMasked[i])

    i = np.argmax(paMin)
    goUp = paMin[i - 2] < paMin[i - 1]  # PA going up at wrap point?

    # Top part
    i0_top = 0 if goUp else i
    i1_top = i if goUp else paMin.size - 1
    paMaxTmp = np.copy(paMax)
    paMaxTmp[np.where(paMin > paMax)[0]] = 360

    # Bottom part
    i = np.argmin(paMax)
    i0_bot = i if goUp else 0
    i1_bot = paMin.size - 1 if goUp else i
    paMinTmp = np.copy(paMin)
    paMinTmp[np.where(paMin > paMax)[0]] = 0

    # Convert datetime to a number for Bokeh
    gdMaskednum = [datetime.date(2019, 6, 1) + datetime.timedelta(days=n)
                   for n, d in enumerate(gdMasked)]
    color = 'green'

    # Draw the curve and error
    try:
        fig.line(gdMaskednum, paMasked, legend_label='cutoff', line_color=color)
    except AttributeError:
        fig.line(gdMaskednum, paMasked, legend='cutoff', line_color=color)

    # Top
    terr_y = np.concatenate([paMin[i0_top:i1_top + 1],
                             paMaxTmp[i0_top:i1_top + 1][::-1]])
    terr_x = np.concatenate([gdMaskednum[i0_top:i1_top + 1],
                             gdMaskednum[i0_top:i1_top + 1][::-1]])
    fig.patch(terr_x, terr_y, color=color, fill_alpha=0.2, line_alpha=0)

    # Bottom
    berr_y = np.concatenate([paMinTmp[i0_bot:i1_bot + 1],
                             paMax[i0_bot:i1_bot + 1][::-1]])
    berr_x = np.concatenate([gdMaskednum[i0_bot:i1_bot + 1],
                             gdMaskednum[i0_bot:i1_bot + 1][::-1]])
    fig.patch(berr_x, berr_y, color='red', fill_alpha=0.2, line_alpha=0)
    from bokeh.plotting import show
    show(fig)

    # Plot formatting
    fig.xaxis.axis_label = 'Date'
    fig.yaxis.axis_label = 'Aperture Position Angle (degrees)'

    return paGood, paBad, gd, fig


def fill_between(fig, xdata, pamin, pamax, **kwargs):
    # addressing NIRSpec issue

    # now creating the patches for the arrays
    nanbot = np.where([np.isnan(i) for i in pamin])[0]
    nantop = np.where([np.isnan(i) for i in pamax])[0]
    yb = np.split(pamin, nanbot)
    xs = np.split(xdata, nanbot)
    yt = np.split(pamax, nantop)
    for x, bot, top in zip(xs, yb, yt):
        x = np.append(x, x[::-1])
        y = np.append(bot, top[::-1])
        fig.patch(x, y, **kwargs)
    return fig


def using_gtvt(
        ra,
        dec,
        instrument,
        targetName='noName',
        ephFileName=None,
        output='bokeh'):
    """Plot the visibility (at a range of position angles) against time.

    Parameters
    ----------
    ra : str
        The RA of the target (in degrees) hh:mm:ss.s or dd:mm:ss.s or representing a float
    dec : str
        The Dec of the target (in degrees) hh:mm:ss.s or dd:mm:ss.s or representing a float
    instrument : str
        Name of the instrument. Can either be (case-sensitive):
        'NIRISS', 'NIRCam', 'MIRI', 'FGS', or 'NIRSpec'
    ephFileName : str
        The filename of the ephemeris file.
    output : str
        Switches on plotting with Bokeh. Parameter value must be 'bokeh'.

    Returns
    -------
    paGood : float
        The good position angle.
    paBad : float
        The bad position angle.
    gd : matplotlib.dates object
       The gregorian date.
    fig : bokeh.plotting.figure object
        The plotted figure.

    """
    # Getting calculations from GTVT (General Target Visibility Tool)
    tab = get_table(ra, dec)

    gd = tab['Date']
    paMin = tab[str(instrument) + ' min']
    paMax = tab[str(instrument) + ' max']
    paNom = tab[str(instrument) + ' nom']
    v3min = tab['V3PA min']
    v3max = tab['V3PA max']

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~NOTE~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Addressing NIRSpec issue*
    # *the issue that NIRSpec's angle goes beyond 360 degrees with some targs,
    # thus resetting back to 0 degrees, which can make the plot look weird
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    index = np.arange(0, len(paNom), 1)

    for idx in index:

        try:
            a1 = paNom[idx]
            b1 = paNom[idx + 1]

            if (np.isfinite(a1)) & (np.isfinite(b1)):
                delta = np.abs(a1 - b1)

                if delta > 250:

                    gd = np.insert(gd, idx + 1, np.nan)
                    paMin = np.insert(paMin, idx + 1, np.nan)
                    paMax = np.insert(paMax, idx + 1, np.nan)
                    paNom = np.insert(paNom, idx + 1, np.nan)
                    v3min = np.insert(v3min, idx + 1, np.nan)
                    v3max = np.insert(v3min, idx + 1, np.nan)
        except BaseException:
            pass

    # Setting up HoverTool parameters & other variables
    COLOR = 'green'
    TOOLS = 'pan, wheel_zoom, box_zoom, reset, save'
    SOURCE = ColumnDataSource(data=dict(pamin=paMin,
                                        panom=paNom,
                                        pamax=paMax,
                                        date=gd))
    TOOLTIPS = [('Date', '@date{%F}'),
                ('Maximum Aperture PA', '@pamax'),
                ('Nominal Aperture PA', '@panom'),
                ('Minimum Aperture PA', '@pamin')]

    # Time to plot
    if output == 'bokeh':
        fig = figure(tools=TOOLS,
                     plot_width=800,
                     plot_height=400,
                     x_axis_type='datetime',
                     title='{} Visibility with {}'.format(targetName,
                                                          instrument))

    # Draw the curve and PA min/max circles
    try:
        nom = fig.line('date', 'panom',
                       line_color=COLOR,
                       legend_label='Nominal Aperture PA',
                       alpha=.5,
                       source=SOURCE)
    except AttributeError:
        nom = fig.line('date', 'panom',
                       line_color=COLOR,
                       legend='Nominal Aperture PA',
                       alpha=.5,
                       source=SOURCE)
                       
    fig.circle('date', 'pamin', color=COLOR, size=1, source=SOURCE)
    fig.circle('date', 'pamax', color=COLOR, size=1, source=SOURCE)

    # Adding HoverTool
    fig.add_tools(HoverTool(renderers=[nom],
                            tooltips=TOOLTIPS,
                            formatters={'date': 'datetime'},
                            mode='vline'))

    # Plot formatting
    fig.xaxis.axis_label = 'Date'
    fig.yaxis.axis_label = 'Aperture Position Angle (degrees)'
    fig.y_range = ranges.Range1d(0, 360)

    # Making the output table
    # Creating new lists w/o the NaN values
    v3minnan, v3maxnan, paNomnan, paMinnan, paMaxnan, gdnan, mjds = \
        [], [], [], [], [], [], []

    for vmin, vmax, pnom, pmin, pmax, date in zip(
            v3min, v3max, paNom, paMin, paMax, gd):
        if np.isfinite(pmin):
            v3minnan.append(vmin)
            v3maxnan.append(vmax)
            paNomnan.append(pnom)
            paMinnan.append(pmin)
            paMaxnan.append(pmax)
            gdnan.append(date)

    # Adding MJD column
    mjdnan = []
    for date in gdnan:
        t = Time(str(date), format='iso')
        mjd = t.mjd
        mjdnan.append(mjd)

    # Adding lists to a table object
    table = Table([v3minnan,
                   v3maxnan,
                   paMinnan,
                   paMaxnan,
                   paNomnan,
                   gdnan,
                   mjdnan],
                  names=('min_V3_PA',
                         'max_V3_PA',
                         'min_Aperture_PA',
                         'max_Aperture_PA',
                         'nom_Aperture_PA',
                         'Gregorian',
                         'MJD'))

    # Getting bad PAs
    allPAs = np.arange(0, 360, 1)
    badPAs = []

    for pa in allPAs:
        if (pa not in np.round(paMinnan)) & \
           (pa not in np.round(paMaxnan)) & \
           (pa not in np.round(paNomnan)):

            badPAs.append(pa)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~NOTE~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # This addresses a bokeh shading issue that accidentally shades
    # accessible PAs (e.g: trappist-1b)
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    remove_pa = []
    for badpa in badPAs:

        for panom in paNomnan:
            diff = np.abs(badpa - panom)
            if diff < 7:
                remove_pa.append(badpa)

    for pa in np.unique(remove_pa):
        badPAs.remove(pa)
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~NOTE~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Grouping the bad PAs into lists within the badPAs list.
    # This will make bad PA shading easier in the contamination Bokeh plot
    # (sossContamFig.py)
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    badPAs = np.sort(badPAs)
    grouped_badPAs = [[badPAs[0]]]

    for idx in range(1, len(badPAs)):

        if ((badPAs[idx - 1] + 1) == badPAs[idx]):

            grouped_badPAs[len(grouped_badPAs) - 1].append(badPAs[idx])

        elif ((badPAs[idx - 1] + 1) < badPAs[idx]):
            grouped_badPAs.append([badPAs[idx]])

    grouped_badPAs = np.asarray(grouped_badPAs)

    return paMin, paMax, gd, fig, table, grouped_badPAs
